#!/usr/bin/env python3
#
#  noble_gases.py
r"""
Group 18: Noble Gases in the Periodic Table.

.. raw:: latex

	\begin{multicols}{2}

.. data:: He

	:class:`~chemistry_tools.elements.classes.Element` representing Helium

.. data:: Ne

	:class:`~chemistry_tools.elements.classes.Element` representing Neon

.. data:: Ar

	:class:`~chemistry_tools.elements.classes.Element` representing Argon

.. data:: Kr

	:class:`~chemistry_tools.elements.classes.Element` representing Krypton

.. data:: Xe

	:class:`~chemistry_tools.elements.classes.Element` representing Xenon

.. data:: Rn

	:class:`~chemistry_tools.elements.classes.Element` representing Radon

.. data:: Og

	:class:`~chemistry_tools.elements.classes.Element` representing Oganesson

.. raw:: latex

	\end{multicols}
"""
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on molmass (https://github.com/cgohlke/molmass)
#  |  Copyright (c) 1990-2020, Christoph Gohlke
#  |  All rights reserved.
#  |  Licensed under the BSD 3-Clause License
#  |  Redistribution and use in source and binary forms, with or without
#  |  modification, are permitted provided that the following conditions are met:
#  |
#  |  1. Redistributions of source code must retain the above copyright notice,
#  |     this list of conditions and the following disclaimer.
#  |
#  |  2. Redistributions in binary form must reproduce the above copyright notice,
#  |     this list of conditions and the following disclaimer in the documentation
#  |     and/or other materials provided with the distribution.
#  |
#  |  3. Neither the name of the copyright holder nor the names of its
#  |     contributors may be used to endorse or promote products derived from
#  |     this software without specific prior written permission.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  |  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  |  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  |  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  |  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  |  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  |  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  |  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  |  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  |  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  |  POSSIBILITY OF SUCH DAMAGE.
#  |
#  Also based on Pyteomics (https://github.com/levitsky/pyteomics)
#  |  Copyright (c) 2011-2015, Anton Goloborodko & Lev Levitsky
#  |  Licensed under the Apache License, Version 2.0 (the "License");
#  |  you may not use this file except in compliance with the License.
#  |  You may obtain a copy of the License at
#  |
#  |    http://www.apache.org/licenses/LICENSE-2.0
#  |
#  |  Unless required by applicable law or agreed to in writing, software
#  |  distributed under the License is distributed on an "AS IS" BASIS,
#  |  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  |  See the License for the specific language governing permissions and
#  |  limitations under the License.
#  |
#  |  See also:
#  |  Goloborodko, A.A.; Levitsky, L.I.; Ivanov, M.V.; and Gorshkov, M.V. (2013)
#  |  "Pyteomics - a Python Framework for Exploratory Data Analysis and Rapid Software
#  |  Prototyping in Proteomics", Journal of The American Society for Mass Spectrometry,
#  |  24(2), 301–304. DOI: `10.1007/s13361-012-0516-6 <http://dx.doi.org/10.1007/s13361-012-0516-6>`_
#  |
#  |  Levitsky, L.I.; Klein, J.; Ivanov, M.V.; and Gorshkov, M.V. (2018)
#  |  "Pyteomics 4.0: five years of development of a Python proteomics framework",
#  |  Journal of Proteome Research.
#  |  DOI: `10.1021/acs.jproteome.8b00717 <http://dx.doi.org/10.1021/acs.jproteome.8b00717>`_
#
#  Isotope data from http://www.nist.gov/pml/data/comp.cfm .

# this package
from .classes import Element

__all__ = ("He", "Ne", "Ar", "Kr", "Xe", "Rn", "Og")

# Isotope 0 Key:
# mass of the most abundant isotope and 1.0 abundance.

He = Element(
		2,
		"He",
		"Helium",
		group=18,
		period=1,
		block='s',
		series=2,
		mass=4.002602,
		eleneg=0.0,
		eleaffin=0.0,
		covrad=0.93,
		atmrad=0.49,
		vdwrad=1.4,
		tboil=4.216,
		tmelt=0.95,
		density=0.1785,
		eleconfig="1s2",
		oxistates='*',
		ionenergy=(24.5874, 54.416),
		isotopes={
				3: (3.0160293191, 1.34e-06),
				4: (4.00260325415, 0.99999866),
				5: (5.01222, 0.0),
				6: (6.0188891, 0.0),
				7: (7.028021, 0.0),
				8: (8.033922, 0.0),
				9: (9.04395, 0.0),
				10: (10.0524, 0.0),
				},
		description=(
				"Colourless, odourless gaseous nonmetallic element. Belongs to "
				"group 18 of the periodic table. Lowest boiling point of all "
				"elements and can only be solidified under pressure. Chemically "
				"inert, no known compounds. Discovered in the solar spectrum in 1868 "
				"by Lockyer."
				)
		)

Ne = Element(
		10,
		"Ne",
		"Neon",
		group=18,
		period=2,
		block='p',
		series=2,
		mass=20.1797,
		eleneg=0.0,
		eleaffin=0.0,
		covrad=0.71,
		atmrad=0.51,
		vdwrad=1.54,
		tboil=27.1,
		tmelt=24.55,
		density=0.8999,
		eleconfig="[He] 2s2 2p6",
		oxistates='*',
		ionenergy=(
				21.5645,
				40.962,
				63.45,
				97.11,
				126.21,
				157.93,
				207.27,
				239.09,
				1195.797,
				1362.164,
				),
		isotopes={
				16: (16.025761, 0.0),
				17: (17.017672, 0.0),
				18: (18.0057082, 0.0),
				19: (19.0018802, 0.0),
				20: (19.9924401754, 0.9048),
				21: (20.99384668, 0.0027),
				22: (21.991385114, 0.0925),
				23: (22.9944669, 0.0),
				24: (23.9936108, 0.0),
				25: (24.997737, 0.0),
				26: (26.000461, 0.0),
				27: (27.00759, 0.0),
				28: (28.01207, 0.0),
				29: (29.01939, 0.0),
				30: (30.0248, 0.0),
				31: (31.03311, 0.0),
				32: (32.04002, 0.0),
				33: (33.04938, 0.0),
				34: (34.05703, 0.0),
				},
		description=(
				"Colourless gaseous element of group 18 on the periodic table "
				"(noble gases). Neon occurs in the atmosphere, and comprises 0.0018% "
				"of the volume of the atmosphere. It has a distinct reddish glow "
				"when used in discharge tubes and neon based lamps. It forms almost "
				"no chemical compounds. Neon was discovered in 1898 by Sir William "
				"Ramsey and M.W. Travers."
				)
		)

Ar = Element(
		18,
		"Ar",
		"Argon",
		group=18,
		period=3,
		block='p',
		series=2,
		mass=39.948,
		eleneg=0.0,
		eleaffin=0.0,
		covrad=0.98,
		atmrad=0.88,
		vdwrad=1.88,
		tboil=87.45,
		tmelt=83.95,
		density=1.66,
		eleconfig="[Ne] 3s2 3p6",
		oxistates='*',
		ionenergy=(
				15.7596,
				27.629,
				40.74,
				59.81,
				75.02,
				91.007,
				124.319,
				143.456,
				422.44,
				478.68,
				538.95,
				618.24,
				686.09,
				755.73,
				854.75,
				918.0,
				4120.778,
				4426.114,
				),
		isotopes={
				30: (30.02156, 0.0),
				31: (31.01212, 0.0),
				32: (31.997638, 0.0),
				33: (32.9899257, 0.0),
				34: (33.9802712, 0.0),
				35: (34.9752576, 0.0),
				36: (35.967545106, 0.003365),
				37: (36.96677632, 0.0),
				38: (37.9627324, 0.000632),
				39: (38.964313, 0.0),
				40: (39.9623831225, 0.996003),
				41: (40.9645006, 0.0),
				42: (41.963046, 0.0),
				43: (42.965636, 0.0),
				44: (43.964924, 0.0),
				45: (44.96804, 0.0),
				46: (45.96809, 0.0),
				47: (46.97219, 0.0),
				48: (47.97454, 0.0),
				49: (48.98052, 0.0),
				50: (49.98443, 0.0),
				51: (50.99163, 0.0),
				52: (51.99678, 0.0),
				53: (53.00494, 0.0),
				},
		description=(
				"Monatomic noble gas. Makes up 0.93% of the air. Colourless, "
				"odorless. Is inert and has no true compounds. Lord Rayleigh and Sir "
				"william Ramsey identified argon in 1894."
				)
		)

Kr = Element(
		36,
		"Kr",
		"Krypton",
		group=18,
		period=4,
		block='p',
		series=2,
		mass=83.798,
		eleneg=0.0,
		eleaffin=0.0,
		covrad=1.12,
		atmrad=1.03,
		vdwrad=2.02,
		tboil=120.85,
		tmelt=116.0,
		density=4.48,
		eleconfig="[Ar] 3d10 4s2 4p6",
		oxistates="2*",
		ionenergy=(
				13.9996,
				24.359,
				36.95,
				52.5,
				64.7,
				78.5,
				110.0,
				126.0,
				230.39,
				),
		isotopes={
				69: (68.96518, 0.0),
				70: (69.95526, 0.0),
				71: (70.94963, 0.0),
				72: (71.942092, 0.0),
				73: (72.939289, 0.0),
				74: (73.9330844, 0.0),
				75: (74.930946, 0.0),
				76: (75.92591, 0.0),
				77: (76.92467, 0.0),
				78: (77.9203648, 0.00355),
				79: (78.920082, 0.0),
				80: (79.916379, 0.02286),
				81: (80.916592, 0.0),
				82: (81.9134836, 0.11593),
				83: (82.914136, 0.115),
				84: (83.911507, 0.56987),
				85: (84.9125273, 0.0),
				86: (85.91061073, 0.17279),
				87: (86.91335486, 0.0),
				88: (87.914447, 0.0),
				89: (88.91763, 0.0),
				90: (89.919517, 0.0),
				91: (90.92345, 0.0),
				92: (91.926156, 0.0),
				93: (92.93127, 0.0),
				94: (93.93436, 0.0),
				95: (94.93984, 0.0),
				96: (95.94307, 0.0),
				97: (96.94856, 0.0),
				98: (97.95191, 0.0),
				99: (98.9576, 0.0),
				100: (99.96114, 0.0),
				},
		description=(
				"Colorless gaseous element, belongs to the noble gases. Occurs in "
				"the air, 0.0001% by volume. It can be extracted from liquid air by "
				"fractional distillation. Generally not isolated, but used with "
				"other inert gases in fluorescent lamps. Five natural isotopes, and "
				"five radioactive isotopes. Kr-85, the most stable radioactive "
				"isotope, has a half-life of 10.76 years and is produced in fission "
				"reactors. Practically inert, though known to form compounds with "
				"Fluorine."
				)
		)

Xe = Element(
		54,
		"Xe",
		"Xenon",
		group=18,
		period=5,
		block='p',
		series=2,
		mass=131.293,
		eleneg=0.0,
		eleaffin=0.0,
		covrad=1.31,
		atmrad=1.24,
		vdwrad=2.16,
		tboil=165.1,
		tmelt=161.39,
		density=4.49,
		eleconfig="[Kr] 4d10 5s2 5p6",
		oxistates="2, 4, 6",
		ionenergy=(12.1298, 21.21, 32.1),
		isotopes={
				110: (109.94428, 0.0),
				111: (110.9416, 0.0),
				112: (111.93562, 0.0),
				113: (112.93334, 0.0),
				114: (113.92798, 0.0),
				115: (114.926294, 0.0),
				116: (115.921581, 0.0),
				117: (116.920359, 0.0),
				118: (117.916179, 0.0),
				119: (118.915411, 0.0),
				120: (119.911784, 0.0),
				121: (120.911462, 0.0),
				122: (121.908368, 0.0),
				123: (122.908482, 0.0),
				124: (123.905893, 0.000952),
				125: (124.9063955, 0.0),
				126: (125.904274, 0.00089),
				127: (126.905184, 0.0),
				128: (127.9035313, 0.019102),
				129: (128.9047794, 0.264006),
				130: (129.903508, 0.04071),
				131: (130.9050824, 0.212324),
				132: (131.9041535, 0.269086),
				133: (132.9059107, 0.0),
				134: (133.9053945, 0.104357),
				135: (134.907227, 0.0),
				136: (135.907219, 0.088573),
				137: (136.911562, 0.0),
				138: (137.91395, 0.0),
				139: (138.918793, 0.0),
				140: (139.92164, 0.0),
				141: (140.92665, 0.0),
				142: (141.92971, 0.0),
				143: (142.93511, 0.0),
				144: (143.93851, 0.0),
				145: (144.94407, 0.0),
				146: (145.94775, 0.0),
				147: (146.95356, 0.0),
				},
		description=(
				"Colourless, odourless gas belonging to group 18 on the periodic "
				"table (the noble gases.) Nine natural isotopes and seven "
				"radioactive isotopes are known. Xenon was part of the first "
				"noble-gas compound synthesized. Several others involving Xenon have "
				"been found since then. Xenon was discovered by Ramsey and Travers "
				"in 1898."
				)
		)

Rn = Element(
		86,
		"Rn",
		"Radon",
		group=18,
		period=6,
		block='p',
		series=2,
		mass=222.0176,
		eleneg=0.0,
		eleaffin=0.0,
		covrad=0.0,
		atmrad=1.34,
		vdwrad=0.0,
		tboil=211.4,
		tmelt=202.0,
		density=9.23,
		eleconfig="[Xe] 4f14 5d10 6s2 6p6",
		oxistates="2*",
		ionenergy=(10.7485, ),
		isotopes={
				195: (195.00544, 0.0),
				196: (196.002115, 0.0),
				197: (197.00158, 0.0),
				198: (197.998679, 0.0),
				199: (198.99837, 0.0),
				200: (199.995699, 0.0),
				201: (200.99563, 0.0),
				202: (201.993263, 0.0),
				203: (202.993387, 0.0),
				204: (203.991429, 0.0),
				205: (204.99172, 0.0),
				206: (205.990214, 0.0),
				207: (206.990734, 0.0),
				208: (207.989642, 0.0),
				209: (208.990415, 0.0),
				210: (209.989696, 0.0),
				211: (210.990601, 0.0),
				212: (211.990704, 0.0),
				213: (212.993883, 0.0),
				214: (213.995363, 0.0),
				215: (214.998745, 0.0),
				216: (216.000274, 0.0),
				217: (217.003928, 0.0),
				218: (218.0056013, 0.0),
				219: (219.0094802, 0.0),
				220: (220.011394, 0.0),
				221: (221.015537, 0.0),
				222: (222.0175777, 0.0),
				223: (223.02179, 0.0),
				224: (224.02409, 0.0),
				225: (225.02844, 0.0),
				226: (226.03089, 0.0),
				227: (227.03541, 0.0),
				228: (228.03799, 0.0),
				},
		description=(
				"Colorless radioactive gaseous element, belongs to the noble gases. "
				"Of the twenty known isotopes, the most stable is Rn-222 with a "
				"half-life of 3.8 days. Formed by the radioactive decay of "
				"Radium-226. Radon itself decays into Polonium. Used in "
				"radiotherapy. As a noble gas, it is effectively inert, though radon "
				"fluoride has been synthesized. First isolated in 1908 by Ramsey and "
				"Gray."
				),
		)

Og = Element(
		118,
		"Og",
		"Oganesson",
		group=18,
		period=7,
		block='p',
		series=8,
		isotopes={
				293: (293.21467, 0.0),
				},
		)



#!/usr/bin/env python
#
#  __init__.py
"""
Python tools for analysis of chemical compounds.
"""
#
#  Copyright (c) 2019-2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Atom, Bond, Compound, Constants, Errors, Lookup and
#  Utils based on PubChemPy by Matt Swain <m.swain@me.com>
#  Available under the MIT License
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation; either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  units, dicts and formulae based on ChemPy (https://github.com/bjodah/chempy)
#  |  Copyright (c) 2015-2018, Björn Dahlgren
#  |  All rights reserved.
#  |
#  |  Redistribution and use in source and binary forms, with or without modification,
#  |  are permitted provided that the following conditions are met:
#  |
#  |    Redistributions of source code must retain the above copyright notice, this
#  |    list of conditions and the following disclaimer.
#  |
#  |    Redistributions in binary form must reproduce the above copyright notice, this
#  |    list of conditions and the following disclaimer in the documentation and/or
#  |    other materials provided with the distribution.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#  |  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#  |  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  |  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
#  |  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#  |  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  |  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
#  |  ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  |  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  |  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#  isotope_masses, linear_regression and formulae based on
#  Pyteomics (https://github.com/levitsky/pyteomics)
#  |  Copyright (c) 2011-2015, Anton Goloborodko & Lev Levitsky
#  |  Licensed under the Apache License, Version 2.0 (the "License");
#  |  you may not use this file except in compliance with the License.
#  |  You may obtain a copy of the License at
#  |
#  |    http://www.apache.org/licenses/LICENSE-2.0
#  |
#  |  Unless required by applicable law or agreed to in writing, software
#  |  distributed under the License is distributed on an "AS IS" BASIS,
#  |  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  |  See the License for the specific language governing permissions and
#  |  limitations under the License.
#  |
#  |  See also:
#  |  Goloborodko, A.A.; Levitsky, L.I.; Ivanov, M.V.; and Gorshkov, M.V. (2013)
#  |  "Pyteomics - a Python Framework for Exploratory Data Analysis and Rapid Software
#  |  Prototyping in Proteomics", Journal of The American Society for Mass Spectrometry,
#  |  24(2), 301–304. DOI: `10.1007/s13361-012-0516-6 <http://dx.doi.org/10.1007/s13361-012-0516-6>`_
#  |
#  |  Levitsky, L.I.; Klein, J.; Ivanov, M.V.; and Gorshkov, M.V. (2018)
#  |  "Pyteomics 4.0: five years of development of a Python proteomics framework",
#  |  Journal of Proteome Research.
#  |  DOI: `10.1021/acs.jproteome.8b00717 <http://dx.doi.org/10.1021/acs.jproteome.8b00717>`_
#

# this package
from chemistry_tools.cache import cached_requests, clear_cache

__author__ = "Dominic Davis-Foster"
__copyright__ = "2019-2020 Dominic Davis-Foster"
__license__ = "LGPLv3+"
__version__ = "0.4.1"
__email__ = "dominic@davis-foster.co.uk"

__all__ = [
		"cached_requests",
		"clear_cache",
		]

if __name__ == "__main__":

	print(__version__)



#  !/usr/bin/env python
#
#  errors.py
"""
Error handling.
"""
#  Copyright (c) 2019-2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation; either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on PubChemPy https://github.com/mcs07/PubChemPy/blob/master/LICENSE
#  |  Copyright 2014 Matt Swain <m.swain@me.com>
#  |  Licensed under the MIT License
#  |
#  |  Permission is hereby granted, free of charge, to any person obtaining a copy
#  |  of this software and associated documentation files (the "Software"), to deal
#  |  in the Software without restriction, including without limitation the rights
#  |  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  |  copies of the Software, and to permit persons to whom the Software is
#  |  furnished to do so, subject to the following conditions:
#
#  |  The above copyright notice and this permission notice shall be included in
#  |  all copies or substantial portions of the Software.
#
#  |  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  |  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  |  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  |  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  |  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  |  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  |  THE SOFTWARE.
#

# stdlib
import json

# 3rd party
import requests

__all__ = [
		"ResponseParseError",
		"PubChemHTTPError",
		"BadRequestError",
		"NotFoundError",
		"MethodNotAllowedError",
		"TimeoutError",
		"HTTPTimeoutError",
		"UnimplementedError",
		"ServerError",
		"HTTP_ERROR_CODES",
		]

#: Numerical list of HTTP status codes considered to be errors.
HTTP_ERROR_CODES = [
		400,  # Bad Request
		401,  # Unauthorized
		402,  # Payment Required
		403,  # Forbidden
		404,  # Not Found
		405,  # Method Not Allowed
		406,  # Not Acceptable
		407,  # Proxy Authentication Required
		408,  # Request Timeout
		409,  # Conflict
		410,  # Gone
		411,  # Length Required
		412,  # Precondition Failed
		413,  # Payload Too Large
		414,  # URI Too Long
		415,  # Unsupported Media Type
		416,  # Range Not Satisfiable
		417,  # Expectation Failed
		421,  # Misdirected Request
		422,  # Unprocessable Entity
		423,  # Locked
		424,  # Failed Dependency
		425,  # Too Early
		426,  # Upgrade Required
		428,  # Precondition Required
		429,  # Too Many Requests
		431,  # Request Header Fields Too Large
		451,  # Unavailable For Legal Reasons
		500,  # Internal Server Error
		501,  # Not Implemented
		502,  # Bad Gateway
		503,  # Service Unavailable
		504,  # Gateway Timeout
		505,  # HTTP Version Not Supported
		506,  # Variant Also Negotiates
		507,  # Insufficient Storage
		508,  # Loop Detected
		509,  # Not Extended
		511,  # Network Authentication Required
		]


class ResponseParseError(Exception):
	"""
	PubChem response is uninterpretable.
	"""


class PubChemHTTPError(Exception):
	"""
	Generic error class to handle all HTTP error codes.
	"""

	def __init__(self, e: requests.Response):
		self.code = e.status_code
		self.msg = e.reason
		try:
			self.msg += f': {json.loads(e.content.decode())["Fault"]["Details"][0]}'
		except (ValueError, IndexError, KeyError):
			pass
		if self.code == 400:
			raise BadRequestError(self.msg)
		elif self.code == 404:
			raise NotFoundError(self.msg)
		elif self.code == 405:
			raise MethodNotAllowedError(self.msg)
		elif self.code == 504:
			raise HTTPTimeoutError(self.msg)
		elif self.code == 501:
			raise UnimplementedError(self.msg)
		elif self.code == 500:
			raise ServerError(self.msg)

	def __str__(self) -> str:
		return repr(self.msg)


class BadRequestError(PubChemHTTPError):
	"""
	Request is improperly formed (syntax error in the URL, POST body, etc.).
	"""

	def __init__(self, msg: str = "Request is improperly formed"):
		self.msg = msg


class NotFoundError(PubChemHTTPError):
	"""
	The input record was not found (e.g. invalid CID).
	"""

	def __init__(self, msg: str = "The input record was not found"):
		self.msg = msg


class MethodNotAllowedError(PubChemHTTPError):
	"""
	Request not allowed (such as invalid MIME type in the HTTP Accept header).
	"""

	def __init__(self, msg: str = "Request not allowed"):
		self.msg = msg


class HTTPTimeoutError(PubChemHTTPError):
	"""
	The request timed out, from server overload or too broad a request.

	.. versionchanged:: 0.4.0  Renamed from TimeoutErrpr
	"""

	def __init__(self, msg: str = "The request timed out"):
		self.msg = msg


TimeoutError = HTTPTimeoutError  # noqa: A001  # pylint: disable=redefined-builtin


class UnimplementedError(PubChemHTTPError):
	"""
	The requested operation has not (yet) been implemented by the server.
	"""

	def __init__(self, msg: str = "The requested operation has not been implemented"):
		self.msg = msg


class ServerError(PubChemHTTPError):
	"""
	Some problem on the server side (such as a database server down, etc.).
	"""

	def __init__(self, msg: str = "Some problem on the server side"):
		self.msg = msg



#!/usr/bin/env python
#
#  bond.py
"""
Represents a bond between atoms in a :class:`~chemistry_tools.pubchem.compound.Compound`.
"""
#
#  Copyright (c) 2019-2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation; either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on PubChemPy https://github.com/mcs07/PubChemPy/blob/master/LICENSE
#  |  Copyright 2014 Matt Swain <m.swain@me.com>
#  |  Licensed under the MIT License
#  |
#  |  Permission is hereby granted, free of charge, to any person obtaining a copy
#  |  of this software and associated documentation files (the "Software"), to deal
#  |  in the Software without restriction, including without limitation the rights
#  |  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  |  copies of the Software, and to permit persons to whom the Software is
#  |  furnished to do so, subject to the following conditions:
#
#  |  The above copyright notice and this permission notice shall be included in
#  |  all copies or substantial portions of the Software.
#
#  |  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  |  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  |  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  |  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  |  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  |  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  |  THE SOFTWARE.
#

# stdlib
from typing import Any, Dict, FrozenSet, Optional, Union

# 3rd party
from domdf_python_tools.doctools import prettify_docstrings
from enum_tools import IntEnum

# this package
from chemistry_tools.pubchem.errors import ResponseParseError

__all__ = ["BondType", "Bond", "parse_bonds"]


class BondType(IntEnum):
	"""
	Enumeration of possible bond types.
	"""

	SINGLE = 1
	DOUBLE = 2
	TRIPLE = 3
	QUADRUPLE = 4
	DATIVE = 5
	COMPLEX = 6
	IONIC = 7
	UNKNOWN = 255


@prettify_docstrings
class Bond:
	"""
	Class to represent a bond between two atoms in a
	:class:`~chemistry_tools.pubchem.compound.Compound`.

	:param aid1: ID of the begin atom of this bond
	:param aid2: ID of the end atom of this bond
	:param order: Bond order
	:param style: Bond style annotation.
	"""  # noqa: D400

	def __init__(
			self,
			aid1: int,
			aid2: int,
			order: Union[int, BondType] = BondType.SINGLE,
			style=None,
			):
		self.aid1: int = aid1
		self.aid2: int = aid2
		self.order: BondType = BondType(order)
		self.style = style

	def __repr__(self) -> str:
		return f"Bond({self.aid1}, {self.aid2}, {self.order!r})"

	def __eq__(self, other) -> bool:  # noqa: MAN001
		return (
			isinstance(other, type(self))
			and self.aid1 == other.aid1
			and self.aid2 == other.aid2
			and self.order == other.order
			and self.style == other.style
			)  # yapf: disable

	def to_dict(self) -> Dict[str, Any]:
		"""
		Return a dictionary containing bond data.

		:rtype:

		.. latex:clearpage::
		"""

		data = {"aid1": self.aid1, "aid2": self.aid2, "order": self.order}

		if self.style is not None:
			data["style"] = self.style

		return data


def parse_bonds(
		bonds_dict: Dict[str, Any],
		coords_dict: Optional[Dict] = None,
		) -> Dict[FrozenSet[int], Bond]:
	"""
	Parse bonds from the given dictionary.

	:param bonds_dict:
	:param coords_dict:
	"""

	bonds: Dict[FrozenSet[int], Bond] = {}

	if not bonds_dict:
		return bonds

	# Create bonds
	aid1s = bonds_dict["aid1"]
	aid2s = bonds_dict["aid2"]
	orders = bonds_dict["order"]

	if not len(aid1s) == len(aid2s) == len(orders):
		raise ResponseParseError("Error parsing bonds")

	for aid1, aid2, order in zip(aid1s, aid2s, orders):
		bonds[frozenset((aid1, aid2))] = Bond(aid1=aid1, aid2=aid2, order=order)

	# Add styles
	if coords_dict and "style" in coords_dict[0]["conformers"][0]:
		aid1s = coords_dict[0]["conformers"][0]["style"]["aid1"]
		aid2s = coords_dict[0]["conformers"][0]["style"]["aid2"]
		styles = coords_dict[0]["conformers"][0]["style"]["annotation"]
		for aid1, aid2, style in zip(aid1s, aid2s, styles):
			bonds[frozenset((aid1, aid2))].style = style

	return bonds



#!/usr/bin/env python3
#
#  species.py
"""
Class to represent a formula with phase information (e.g. solid, liquid, gas, or aqueous).
"""
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on ChemPy (https://github.com/bjodah/chempy)
#  |  Copyright (c) 2015-2018, Björn Dahlgren
#  |  All rights reserved.
#  |
#  |  Redistribution and use in source and binary forms, with or without modification,
#  |  are permitted provided that the following conditions are met:
#  |
#  |    Redistributions of source code must retain the above copyright notice, this
#  |    list of conditions and the following disclaimer.
#  |
#  |    Redistributions in binary form must reproduce the above copyright notice, this
#  |    list of conditions and the following disclaimer in the documentation and/or
#  |    other materials provided with the distribution.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#  |  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#  |  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  |  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
#  |  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#  |  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  |  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
#  |  ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  |  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  |  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#  Also based on Pyteomics (https://github.com/levitsky/pyteomics)
#  |  Copyright (c) 2011-2015, Anton Goloborodko & Lev Levitsky
#  |  Licensed under the Apache License, Version 2.0 (the "License");
#  |  you may not use this file except in compliance with the License.
#  |  You may obtain a copy of the License at
#  |
#  |    http://www.apache.org/licenses/LICENSE-2.0
#  |
#  |  Unless required by applicable law or agreed to in writing, software
#  |  distributed under the License is distributed on an "AS IS" BASIS,
#  |  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  |  See the License for the specific language governing permissions and
#  |  limitations under the License.
#  |
#  |  See also:
#  |  Goloborodko, A.A.; Levitsky, L.I.; Ivanov, M.V.; and Gorshkov, M.V. (2013)
#  |  "Pyteomics - a Python Framework for Exploratory Data Analysis and Rapid Software
#  |  Prototyping in Proteomics", Journal of The American Society for Mass Spectrometry,
#  |  24(2), 301–304. DOI: `10.1007/s13361-012-0516-6 <http://dx.doi.org/10.1007/s13361-012-0516-6>`_
#  |
#  |  Levitsky, L.I.; Klein, J.; Ivanov, M.V.; and Gorshkov, M.V. (2018)
#  |  "Pyteomics 4.0: five years of development of a Python proteomics framework",
#  |  Journal of Proteome Research.
#  |  DOI: `10.1021/acs.jproteome.8b00717 <http://dx.doi.org/10.1021/acs.jproteome.8b00717>`_
#
#  Also based on molmass (https://github.com/cgohlke/molmass)
#  |  Copyright (c) 1990-2020, Christoph Gohlke
#  |  All rights reserved.
#  |  Licensed under the BSD 3-Clause License
#  |  Redistribution and use in source and binary forms, with or without
#  |  modification, are permitted provided that the following conditions are met:
#  |
#  |  1. Redistributions of source code must retain the above copyright notice,
#  |     this list of conditions and the following disclaimer.
#  |
#  |  2. Redistributions in binary form must reproduce the above copyright notice,
#  |     this list of conditions and the following disclaimer in the documentation
#  |     and/or other materials provided with the distribution.
#  |
#  |  3. Neither the name of the copyright holder nor the names of its
#  |     contributors may be used to endorse or promote products derived from
#  |     this software without specific prior written permission.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  |  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  |  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  |  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  |  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  |  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  |  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  |  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  |  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  |  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  |  POSSIBILITY OF SUCH DAMAGE.
#  |

# stdlib
from typing import Dict, List, Optional, Type, TypeVar, cast

# 3rd party
from cawdrey import frozendict  # nodep
from typing_extensions import Literal  # nodep

# this package
from chemistry_tools.formulae.formula import Formula

__all__ = ["Species", 'S']

S = TypeVar('S', bound="Species")


class Species(Formula):
	"""
	Formula with phase information (e.g. solid, liquid, gas, or aqueous).

	Species extends :class:`~chemistry_tools.formulae.formula.Formula` with the new attribute :attr:`~.phase`

	:param composition: A :class:`~chemistry_tools.formulae.formula.Formula` object with the elemental
		composition of a substance, or a :class:`python:dict` representing the same.
		If :py:obj:`None` an empty object is created
	:param charge:
	:param phase: Either ``'s'``, ``'l'``, ``'g'``, or ``'aq'``. :py:obj:`None` represents an unknown phase.
	"""

	_phases: frozendict[str, str] = frozendict(s="Solid", l="Liquid", g="Gas", aq="Aqueous")

	#: The phase of the species (e.g. solid, liquid, gas, or aqueous). :py:obj:`None` represents an unknown phase.
	phase: Optional[Literal["s", "l", "g", "aq"]]

	def __init__(
			self,
			composition: Optional[Dict[str, int]] = None,
			charge: int = 0,
			phase: Optional[Literal['s', 'l', 'g', "aq"]] = None,
			):
		super().__init__(composition, charge)

		self.phase = phase

	@classmethod
	def from_kwargs(
			cls: Type['S'],
			*,
			charge: int = 0,
			phase: Optional[Literal['s', 'l', 'g', "aq"]] = None,
			**kwargs
			) -> S:
		"""
		Create a new :class:`~chemistry_tools.formulae.species.Species` object from keyword
		arguments representing the elements in the compound.

		:param charge: The charge of the compound
		:param phase: The phase of the compound (e.g. ``'s'`` for solid)
		"""  # noqa: D400

		return cls(kwargs, charge=charge, phase=phase)

	@classmethod
	def from_string(
			cls: Type['S'],
			formula: str,
			charge: int = 0,
			phase: Optional[Literal['s', 'l', 'g', "aq"]] = None,
			) -> S:
		"""
		Create a new :class:`~chemistry_tools.formulae.species.Species` object by parsing a string.

		.. note:: Isotopes cannot (currently) be parsed using this method

		:param formula: A string with a chemical formula
		:param phase: Either ``'s'``, ``'l'``, ``'g'``, or ``'aq'``. :py:obj:`None` represents an unknown phase.
		:param charge:

		:rtype:

		.. latex:clearpage::

		:bold-title:`Examples:`

		.. code-block:: python

			>>> water = Species.from_string('H2O')
			>>> water.phase
			None
			>>> NaCl = Species.from_string('NaCl(s)')
			>>> NaCl.phase
			s
			>>> Hg_l = Species.from_string('Hg(l)')
			>>> Hg_l.phase
			l
			>>> CO2g = Species.from_string('CO2(g)')
			>>> CO2g.phase
			g
			>>> CO2aq = Species.from_string('CO2(aq)')
			>>> CO2aq.phase
			aq
		"""

		if phase is None:
			for p in cls._phases:
				if formula.endswith(f"({p})"):
					phase = cast(Optional[Literal['s', 'l', 'g', "aq"]], p)
					break

		f = super().from_string(formula, charge)
		f.phase = phase

		return f

	def copy(self: S) -> S:
		"""
		Returns a copy of the :class:`~.Species`.
		"""

		return self.__class__(
				self,
				charge=self.charge,
				phase=cast(Optional[Literal['s', 'l', 'g', "aq"]], self.phase),
				)

	def __eq__(self, other) -> bool:  # noqa: MAN001
		"""
		Returns ``self == other``.
		"""

		if isinstance(other, Species):
			if super().__eq__(other):
				return self.phase == other.phase

		return super().__eq__(other)

	def _repr_elements(self) -> List[str]:
		elements = super()._repr_elements()

		if self.phase:
			elements.append(f"phase={self.phase}")

		return elements

	@property
	def hill_formula(self) -> str:
		"""
		Returns the formula in Hill notation.

		:bold-title:`Examples:`

		.. code-block:: python

			>>> Species.from_string('BrC2H5').hill_formula
			'C2H5Br'
			>>> Species.from_string('HBr').hill_formula
			'BrH'
			>>> Species.from_string('[(CH3)3Si2]2NNa').hill_formula
			'C6H18NNaSi4'
		"""

		hill = super().hill_formula

		if self.phase:
			return f"{hill}({self.phase})"
		else:
			return hill

	@property
	def empirical_formula(self) -> str:
		"""
		Returns the empirical formula in Hill notation.

		The empirical formula has the simplest whole number ratio of atoms
		of each element present in the formula.

		:bold-title:`Examples:`

		.. code-block:: python

			>>> Formula.from_string('H2O').empirical_formula
			'H2O'
			>>> Formula.from_string('S4').empirical_formula
			'S'
			>>> Formula.from_string('C6H12O6').empirical_formula
			'CH2O'
		"""

		hill = super().empirical_formula

		if self.phase:
			return f"{hill}({self.phase})"
		else:
			return hill



#!/usr/bin/env python3
#
#  alkaline_earth_metals.py
r"""
Group 2: Alkaline Earth Metals in the Periodic Table.

.. raw:: latex

	\begin{multicols}{2}

.. data:: Be

	:class:`~chemistry_tools.elements.classes.Element` representing Beryllium

.. data:: Mg

	:class:`~chemistry_tools.elements.classes.Element` representing Magnesium

.. data:: Ca

	:class:`~chemistry_tools.elements.classes.Element` representing Calcium

.. data:: Sr

	:class:`~chemistry_tools.elements.classes.Element` representing Strontium

.. data:: Ba

	:class:`~chemistry_tools.elements.classes.Element` representing Barium

.. data:: Ra

	:class:`~chemistry_tools.elements.classes.Element` representing Radium

.. raw:: latex

	\end{multicols}
"""
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on molmass (https://github.com/cgohlke/molmass)
#  |  Copyright (c) 1990-2020, Christoph Gohlke
#  |  All rights reserved.
#  |  Licensed under the BSD 3-Clause License
#  |  Redistribution and use in source and binary forms, with or without
#  |  modification, are permitted provided that the following conditions are met:
#  |
#  |  1. Redistributions of source code must retain the above copyright notice,
#  |     this list of conditions and the following disclaimer.
#  |
#  |  2. Redistributions in binary form must reproduce the above copyright notice,
#  |     this list of conditions and the following disclaimer in the documentation
#  |     and/or other materials provided with the distribution.
#  |
#  |  3. Neither the name of the copyright holder nor the names of its
#  |     contributors may be used to endorse or promote products derived from
#  |     this software without specific prior written permission.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  |  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  |  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  |  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  |  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  |  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  |  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  |  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  |  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  |  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  |  POSSIBILITY OF SUCH DAMAGE.
#  |
#  Also based on Pyteomics (https://github.com/levitsky/pyteomics)
#  |  Copyright (c) 2011-2015, Anton Goloborodko & Lev Levitsky
#  |  Licensed under the Apache License, Version 2.0 (the "License");
#  |  you may not use this file except in compliance with the License.
#  |  You may obtain a copy of the License at
#  |
#  |    http://www.apache.org/licenses/LICENSE-2.0
#  |
#  |  Unless required by applicable law or agreed to in writing, software
#  |  distributed under the License is distributed on an "AS IS" BASIS,
#  |  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  |  See the License for the specific language governing permissions and
#  |  limitations under the License.
#  |
#  |  See also:
#  |  Goloborodko, A.A.; Levitsky, L.I.; Ivanov, M.V.; and Gorshkov, M.V. (2013)
#  |  "Pyteomics - a Python Framework for Exploratory Data Analysis and Rapid Software
#  |  Prototyping in Proteomics", Journal of The American Society for Mass Spectrometry,
#  |  24(2), 301–304. DOI: `10.1007/s13361-012-0516-6 <http://dx.doi.org/10.1007/s13361-012-0516-6>`_
#  |
#  |  Levitsky, L.I.; Klein, J.; Ivanov, M.V.; and Gorshkov, M.V. (2018)
#  |  "Pyteomics 4.0: five years of development of a Python proteomics framework",
#  |  Journal of Proteome Research.
#  |  DOI: `10.1021/acs.jproteome.8b00717 <http://dx.doi.org/10.1021/acs.jproteome.8b00717>`_
#
#  Isotope data from http://www.nist.gov/pml/data/comp.cfm .

# this package
from .classes import Element

__all__ = ("Be", "Mg", "Ca", "Sr", "Ba", "Ra")

# Isotope 0 Key:
# mass of the most abundant isotope and 1.0 abundance.

Be = Element(
		4,
		"Be",
		"Beryllium",
		group=2,
		period=2,
		block='s',
		series=4,
		mass=9.0121831,
		eleneg=1.57,
		eleaffin=0.0,
		covrad=0.9,
		atmrad=1.4,
		vdwrad=0.0,
		tboil=3243.0,
		tmelt=1560.0,
		density=1.85,
		eleconfig="[He] 2s2",
		oxistates="2*",
		ionenergy=(9.3227, 18.211, 153.893, 217.713),
		isotopes={
				5: (5.04079, 0.0),
				6: (6.019726, 0.0),
				7: (7.01692983, 0.0),
				8: (8.0053051, 0.0),
				9: (9.0121822, 1.0),
				10: (10.0135338, 0.0),
				11: (11.021658, 0.0),
				12: (12.026921, 0.0),
				13: (13.03569, 0.0),
				14: (14.04289, 0.0),
				15: (15.05346, 0.0),
				16: (16.06192, 0.0),
				},
		description=(
				"Grey metallic element of group 2 of the periodic table. Is toxic "
				"and can cause severe lung diseases and dermatitis. Shows high "
				"covalent character. It was isolated independently by F. Wohler and "
				"A.A. Bussy in 1828."
				)
		)

Mg = Element(
		12,
		"Mg",
		"Magnesium",
		group=2,
		period=3,
		block='s',
		series=4,
		mass=24.3051,
		eleneg=1.31,
		eleaffin=0.0,
		covrad=1.36,
		atmrad=1.72,
		vdwrad=1.73,
		tboil=1380.0,
		tmelt=922.0,
		density=1.74,
		eleconfig="[Ne] 3s2",
		oxistates="2*",
		ionenergy=(
				7.6462,
				15.035,
				80.143,
				109.24,
				141.26,
				186.5,
				224.94,
				265.9,
				327.95,
				367.53,
				1761.802,
				1962.613,
				),
		isotopes={
				19: (19.03547, 0.0),
				20: (20.018863, 0.0),
				21: (21.011713, 0.0),
				22: (21.9995738, 0.0),
				23: (22.9941237, 0.0),
				24: (23.9850417, 0.7899),
				25: (24.98583692, 0.1),
				26: (25.982592929, 0.1101),
				27: (26.98434059, 0.0),
				28: (27.9838768, 0.0),
				29: (28.9886, 0.0),
				30: (29.990434, 0.0),
				31: (30.996546, 0.0),
				32: (31.998975, 0.0),
				33: (33.005254, 0.0),
				34: (34.00946, 0.0),
				35: (35.01734, 0.0),
				36: (36.023, 0.0),
				37: (37.0314, 0.0),
				38: (38.03757, 0.0),
				39: (39.04677, 0.0),
				40: (40.05393, 0.0),
				},
		description=(
				"Silvery metallic element belonging to group 2 of the periodic "
				"table (alkaline-earth metals). It is essential for living "
				"organisms, and is used in a number of light alloys. Chemically very "
				"reactive, it forms a protective oxide coating when exposed to air "
				"and burns with an intense white flame. It also reacts with sulphur, "
				"nitrogen and the halogens. First isolated by Bussy in 1828."
				)
		)

Ca = Element(
		20,
		"Ca",
		"Calcium",
		group=2,
		period=4,
		block='s',
		series=4,
		mass=40.078,
		eleneg=1.0,
		eleaffin=0.02455,
		covrad=1.74,
		atmrad=2.23,
		vdwrad=0.0,
		tboil=1757.0,
		tmelt=1112.0,
		density=1.54,
		eleconfig="[Ar] 4s2",
		oxistates="2*",
		ionenergy=(
				6.1132,
				11.71,
				50.908,
				67.1,
				84.41,
				108.78,
				127.7,
				147.24,
				188.54,
				211.27,
				591.25,
				656.39,
				726.03,
				816.61,
				895.12,
				974.0,
				1087.0,
				1157.0,
				5129.045,
				5469.738,
				),
		isotopes={
				34: (34.01412, 0.0),
				35: (35.00494, 0.0),
				36: (35.99309, 0.0),
				37: (36.98587, 0.0),
				38: (37.976318, 0.0),
				39: (38.9707197, 0.0),
				40: (39.96259098, 0.96941),
				41: (40.96227806, 0.0),
				42: (41.95861801, 0.00647),
				43: (42.9587666, 0.00135),
				44: (43.9554818, 0.02086),
				45: (44.9561866, 0.0),
				46: (45.9536926, 4e-05),
				47: (46.954546, 0.0),
				48: (47.952534, 0.00187),
				49: (48.955674, 0.0),
				50: (49.957519, 0.0),
				51: (50.9615, 0.0),
				52: (51.9651, 0.0),
				53: (52.97005, 0.0),
				54: (53.97435, 0.0),
				55: (54.98055, 0.0),
				56: (55.98557, 0.0),
				57: (56.99236, 0.0),
				},
		description=(
				"Soft grey metallic element belonging to group 2 of the periodic "
				"table. Used a reducing agent in the extraction of thorium, "
				"zirconium and uranium. Essential element for living organisms."
				)
		)

Sr = Element(
		38,
		"Sr",
		"Strontium",
		group=2,
		period=5,
		block='s',
		series=4,
		mass=87.62,
		eleneg=0.95,
		eleaffin=0.05206,
		covrad=1.91,
		atmrad=2.45,
		vdwrad=0.0,
		tboil=1655.0,
		tmelt=1042.0,
		density=2.63,
		eleconfig="[Kr] 5s2",
		oxistates="2*",
		ionenergy=(
				5.6949,
				11.03,
				43.6,
				57.0,
				71.6,
				90.8,
				106.0,
				122.3,
				162.0,
				177.0,
				324.1,
				),
		isotopes={
				73: (72.96597, 0.0),
				74: (73.95631, 0.0),
				75: (74.94995, 0.0),
				76: (75.94177, 0.0),
				77: (76.937945, 0.0),
				78: (77.93218, 0.0),
				79: (78.929708, 0.0),
				80: (79.924521, 0.0),
				81: (80.923212, 0.0),
				82: (81.918402, 0.0),
				83: (82.917557, 0.0),
				84: (83.913425, 0.0056),
				85: (84.912933, 0.0),
				86: (85.9092602, 0.0986),
				87: (86.9088771, 0.07),
				88: (87.9056121, 0.8258),
				89: (88.9074507, 0.0),
				90: (89.907738, 0.0),
				91: (90.910203, 0.0),
				92: (91.911038, 0.0),
				93: (92.914026, 0.0),
				94: (93.915361, 0.0),
				95: (94.919359, 0.0),
				96: (95.921697, 0.0),
				97: (96.926153, 0.0),
				98: (97.928453, 0.0),
				99: (98.93324, 0.0),
				100: (99.93535, 0.0),
				101: (100.94052, 0.0),
				102: (101.94302, 0.0),
				103: (102.94895, 0.0),
				104: (103.95233, 0.0),
				105: (104.95858, 0.0),
				},
		description=(
				"Soft yellowish metallic element, belongs to group 2 of the "
				"periodic table. Highly reactive chemically. Sr-90 is present in "
				"radioactive fallout and has a half-life of 28 years. Discovered in "
				"1798 by Klaproth and Hope, isolated in 1808 by Humphry Davy."
				)
		)

Ba = Element(
		56,
		"Ba",
		"Barium",
		group=2,
		period=6,
		block='s',
		series=4,
		mass=137.327,
		eleneg=0.89,
		eleaffin=0.14462,
		covrad=1.98,
		atmrad=2.78,
		vdwrad=0.0,
		tboil=2078.0,
		tmelt=1002.0,
		density=3.65,
		eleconfig="[Xe] 6s2",
		oxistates="2*",
		ionenergy=(5.2117, 100.004),
		isotopes={
				114: (113.95068, 0.0),
				115: (114.94737, 0.0),
				116: (115.94138, 0.0),
				117: (116.9385, 0.0),
				118: (117.93304, 0.0),
				119: (118.93066, 0.0),
				120: (119.92604, 0.0),
				121: (120.92405, 0.0),
				122: (121.9199, 0.0),
				123: (122.918781, 0.0),
				124: (123.915094, 0.0),
				125: (124.914473, 0.0),
				126: (125.91125, 0.0),
				127: (126.911094, 0.0),
				128: (127.908318, 0.0),
				129: (128.908679, 0.0),
				130: (129.9063208, 0.00106),
				131: (130.906941, 0.0),
				132: (131.9050613, 0.00101),
				133: (132.9060075, 0.0),
				134: (133.9045084, 0.02417),
				135: (134.9056886, 0.06592),
				136: (135.9045759, 0.07854),
				137: (136.9058274, 0.11232),
				138: (137.9052472, 0.71698),
				139: (138.9088413, 0.0),
				140: (139.910605, 0.0),
				141: (140.914411, 0.0),
				142: (141.916453, 0.0),
				143: (142.920627, 0.0),
				144: (143.922953, 0.0),
				145: (144.92763, 0.0),
				146: (145.93022, 0.0),
				147: (146.93495, 0.0),
				148: (147.93772, 0.0),
				149: (148.94258, 0.0),
				150: (149.94568, 0.0),
				151: (150.95081, 0.0),
				152: (151.95427, 0.0),
				153: (152.95961, 0.0),
				},
		description=(
				"Silvery-white reactive element, belonging to group 2 of the "
				"periodic table. Soluble barium compounds are extremely poisonous. "
				"Identified in 1774 by Karl Scheele and extracted in 1808 by Humphry "
				"Davy."
				),
		)

Ra = Element(
		88,
		"Ra",
		"Radium",
		group=2,
		period=7,
		block='s',
		series=4,
		mass=226.0254,
		eleneg=0.9,
		eleaffin=0.0,
		covrad=0.0,
		atmrad=0.0,
		vdwrad=0.0,
		tboil=1413.0,
		tmelt=973.0,
		density=5.5,
		eleconfig="[Rn] 7s2",
		oxistates="2*",
		ionenergy=(5.2784, 10.147),
		isotopes={
				202: (202.00989, 0.0),
				203: (203.00927, 0.0),
				204: (204.0065, 0.0),
				205: (205.00627, 0.0),
				206: (206.003827, 0.0),
				207: (207.0038, 0.0),
				208: (208.00184, 0.0),
				209: (209.00199, 0.0),
				210: (210.000495, 0.0),
				211: (211.000898, 0.0),
				212: (211.999794, 0.0),
				213: (213.000384, 0.0),
				214: (214.000108, 0.0),
				215: (215.00272, 0.0),
				216: (216.003533, 0.0),
				217: (217.00632, 0.0),
				218: (218.00714, 0.0),
				219: (219.010085, 0.0),
				220: (220.011028, 0.0),
				221: (221.013917, 0.0),
				222: (222.015375, 0.0),
				223: (223.0185022, 0.0),
				224: (224.0202118, 0.0),
				225: (225.023612, 0.0),
				226: (226.0254098, 0.0),
				227: (227.0291778, 0.0),
				228: (228.0310703, 0.0),
				229: (229.034958, 0.0),
				230: (230.037056, 0.0),
				231: (231.04122, 0.0),
				232: (232.04364, 0.0),
				233: (233.04806, 0.0),
				234: (234.0507, 0.0),
				},
		description=(
				"Radioactive metallic transuranic element, belongs to group 2 of "
				"the periodic table. Most stable isotope, Ra-226 has a half-life of "
				"1602 years, which decays into radon. Isolated from pitchblende in "
				"1898 Marie and Pierre Curie."
				)
		)



#  !/usr/bin/env python
#
#  lookup.py
"""
Lookup properties for compound by name or CAS number.
"""
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation; either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

# stdlib
from typing import List, Sequence, Union

# this package
from chemistry_tools.pubchem.compound import Compound
from chemistry_tools.pubchem.description import parse_description, rest_get_description
from chemistry_tools.pubchem.enums import PubChemNamespace

__all__ = ["get_compounds"]

# TODO: xrefs
# TODO: formula search with listkey and pagination 	https://pubchemdocs.ncbi.nlm.nih.gov/pug-rest$_Toc494865589


def get_compounds(
		identifier: Union[str, int, Sequence[Union[str, int]]],
		namespace: Union[PubChemNamespace, str] = PubChemNamespace.name,
		) -> List[Compound]:
	"""
	Returns a list of Compound objects for compounds that match the search criteria.

	As more than one compound may be identified the results are returned in a list.

	:param identifier: Identifiers (e.g. name, CID) for the compound to look up.
		When using the CID namespace data for multiple compounds can be retrieved at once by
		supplying either a comma-separated string or a list.
	:param namespace: The type of identifier to look up. Valid values are in :class:`~.PubChemNamespace`.
	"""

	data = rest_get_description(identifier, namespace)

	compounds = []

	for record in parse_description(data):
		compounds.append(Compound(record["Title"], record["CID"], record["Description"]))

	return compounds



#!/usr/bin/env python3
#
#  synonyms.py
"""
Functions for obtaining the synonyms of a compound from the PubChem database.
"""
#
#  Copyright (c) 2019-2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation; either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

# stdlib
from typing import Dict, List, Sequence, Union

# this package
from chemistry_tools.pubchem.enums import PubChemNamespace
from chemistry_tools.pubchem.pug_rest import do_rest_get

__all__ = ["Synonyms", "get_synonyms", "rest_get_synonyms"]


class Synonyms(List[str]):
	"""
	Contains a list of synonyms for a compound.

	:param initlist: The content to initialise the list with.
	"""

	def __init__(self, initlist):  # noqa: MAN001
		super().__init__()

		for val in initlist:
			self.append(str(val))

	def __contains__(self, synonym) -> bool:  # noqa: MAN001
		"""
		Return ``synonym in self``.

		The comparison treats hyphens and underscores as whitespace.

		:param synonym:
		"""

		for v in self:
			if self._prep_contains(v) == self._prep_contains(synonym):
				return True
		return False

	def append(self, synonym: str) -> None:
		"""
		Append ``synonym`` to the end of the list.

		:param synonym:
		"""

		if synonym not in self:
			super().append(str(synonym))

	@staticmethod
	def _prep_contains(val: str) -> str:
		val = val.casefold()

		for remove in ['-', '_', ' ']:
			val = val.replace(remove, ' ')

		return val


def get_synonyms(
		identifier: Union[str, int, Sequence[Union[str, int]]],
		namespace: Union[PubChemNamespace, str] = PubChemNamespace.name,
		) -> List[Dict]:
	"""
	Returns a list of synonyms for the compound with the given identifier.
	As more than one compound may be identified the results are returned in a list.

	:param identifier: Identifiers (e.g. name, CID) for the compound to look up.
		When using the CID namespace data for multiple compounds can be retrieved at once by
		supplying either a comma-separated string or a list.
	:param namespace: The type of identifier to look up.
		Valid values are in :class:`~.PubChemNamespace`.

	:return: List of dictionaries containing the CID and a list of synonyms for the compounds.
	"""

	data = rest_get_synonyms(identifier, namespace)

	results = []

	for compound in data["InformationList"]["Information"]:
		parsed_data = {
				"CID": compound["CID"],
				"synonyms": Synonyms(compound["Synonym"][:20]),
				}

		results.append(parsed_data)

	return results


def rest_get_synonyms(
		identifier: Union[str, int, Sequence[Union[str, int]]],
		namespace: Union[PubChemNamespace, str] = PubChemNamespace.name,
		**kwargs,
		) -> Dict:
	"""
	Get the list of synonyms for the given compound.

	:param identifier: Identifiers (e.g. name, CID) for the compound to look up.
		When using the CID namespace data for multiple compounds can be retrieved at once by
		supplying either a comma-separated string or a list.
	:param namespace: The type of identifier to look up. Valid values are in :class:`~.PubChemNamespace`
	:param kwargs: Optional arguments that ``json.loads`` takes.

	:raises ValueError: If the response body does not contain valid JSON.

	:return: Parsed JSON data.
	"""

	return do_rest_get(namespace, identifier, domain="synonyms").json(**kwargs)



#!/usr/bin/env python3
#
#  images.py
"""
Functions for handling images.
"""
#
#  Copyright (c) 2019-2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation; either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

# stdlib
from io import BytesIO
from typing import Sequence, Union

# 3rd party
from PIL import Image  # type: ignore[import]  # nodep

# this package
from .enums import PubChemNamespace
from .pug_rest import do_rest_get

__all__ = ["get_structure_image"]


def get_structure_image(
		identifier: Union[str, int, Sequence[Union[str, int]]],
		namespace: Union[PubChemNamespace, str] = PubChemNamespace.name,
		width: int = 300,
		height: int = 300
		) -> Image.Image:
	"""
	Returns an image of the structure of the compound with the given name.

	:param identifier: Identifiers (e.g. name, CID) for the compound to look up.
		When using the CID namespace data for multiple compounds can be retrieved at once by
		supplying either a comma-separated string or a list.
	:param namespace: The type of identifier to look up. Valid values are in :class:`~.PubChemNamespace`.
	:param width: The image width in pixels.
	:param height: The image height in pixels.

	:return: Pillow Image data
	"""

	r = do_rest_get(namespace, identifier, "PNG", png_width=width, png_height=height)
	return Image.open(BytesIO(r.content))


rest_get_structure_image = get_structure_image



#!/usr/bin/env python3
#
#  elements.py
"""
Provides classes to model period table elements.
"""
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on molmass (https://github.com/cgohlke/molmass)
#  |  Copyright (c) 1990-2020, Christoph Gohlke
#  |  All rights reserved.
#  |  Licensed under the BSD 3-Clause License
#  |  Redistribution and use in source and binary forms, with or without
#  |  modification, are permitted provided that the following conditions are met:
#  |
#  |  1. Redistributions of source code must retain the above copyright notice,
#  |     this list of conditions and the following disclaimer.
#  |
#  |  2. Redistributions in binary form must reproduce the above copyright notice,
#  |     this list of conditions and the following disclaimer in the documentation
#  |     and/or other materials provided with the distribution.
#  |
#  |  3. Neither the name of the copyright holder nor the names of its
#  |     contributors may be used to endorse or promote products derived from
#  |     this software without specific prior written permission.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  |  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  |  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  |  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  |  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  |  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  |  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  |  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  |  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  |  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  |  POSSIBILITY OF SUCH DAMAGE.
#  |
#

# stdlib
import functools
from typing import Dict, Iterable, Iterator, List, Optional, Tuple, Union, overload

# 3rd party
from domdf_python_tools import doctools
from domdf_python_tools.bases import Dictable
from domdf_python_tools.doctools import prettify_docstrings

# this package
from chemistry_tools._memoized_property import memoized_property

# this package
from . import _elements, _table

__all__ = ["Element", "Isotope", "Elements", "HeavyHydrogen", "IsotopeDict"]

#: Type alias for isotope dictionaries.
IsotopeDict = Dict[int, Union["Isotope", Tuple[float, float]]]


@prettify_docstrings
class Element(Dictable):
	"""
	Chemical element.

	:param number: The atomic number of the element.
	:param symbol: The chemical symbol of the element.
	:param name: The name of the element in English.
	:param group: The number of electrons in the element.
	:param period: The number of protons in the element.
	:param block: The group of the element in the periodic table.
	:param series: The Period of the element in the periodic table.
	:param mass: The relative atomic mass.
	:param eleneg: The Electronegativity (Pauling scale).
	:param eleaffin: The electron affinity in eV.
	:param covrad: The Covalent radius in Angstrom.
	:param atmrad: The Atomic radius in Angstrom.
	:param vdwrad: The Van der Waals radius in Angstrom.
	:param tboil: The boiling temperature in K.
	:param tmelt: The melting temperature in K.
	:param density: The density at 295K in g/cm³ respectively g/L.
	:param eleconfig: The Ground state electron configuration.
	:param oxistates: The oxidation states.
	:param ionenergy: The ionization energies in ``eV``.
	:param isotopes: The Isotopic composition. A mapping of isotope mass numbers to :class:`~.Isotope` objects.
	:param description: A description of the element.

	.. autosummary-widths:: 30/100
	"""

	_ionenergy: Tuple

	def __init__(
			self,
			number: int,
			symbol: str,
			name: str,
			group: int = 0,
			period: int = 0,
			block: str = '',
			series: int = 0,
			mass: float = 0.0,
			eleneg: float = 0.0,
			eleaffin: float = 0.0,
			covrad: float = 0.0,
			atmrad: float = 0.0,
			vdwrad: float = 0.0,
			tboil: float = 0.0,
			tmelt: float = 0.0,
			density: float = 0.0,
			eleconfig: str = '',
			oxistates: str = '',
			ionenergy: Optional[Tuple] = None,  # TODO
			isotopes: Optional[IsotopeDict] = None,
			description: str = '',
			):

		super().__init__()

		self._number: int = number
		self._symbol: str = symbol
		self._name: str = name
		self._electrons: int = number
		self._protons: int = number
		self._group: int = group
		self._period: int = period
		self._block: str = block
		self._series: int = series
		self._mass: float = mass
		self._eleneg: float = eleneg
		self._eleaffin: float = eleaffin
		self._covrad: float = covrad
		self._atmrad: float = atmrad
		self._vdwrad: float = vdwrad
		self._tboil: float = tboil
		self._tmelt: float = tmelt
		self._density: float = density
		self._eleconfig: str = eleconfig
		self._oxistates: str = oxistates
		self._description: str = description

		if ionenergy is None:
			self._ionenergy = tuple()
		else:
			self._ionenergy = ionenergy

		self._isotopes: Dict[int, Isotope] = {}

		if isotopes is not None:
			for massnumber, isotope in isotopes.items():
				if isinstance(isotope, Isotope):
					self._isotopes[int(massnumber)] = isotope
				elif isinstance(isotope, (list, tuple)):
					self._isotopes[int(massnumber)] = Isotope(*isotope, massnumber)

	@property
	def __dict__(self):  # noqa: MAN002
		return dict(
				number=self._number,
				symbol=self._symbol,
				name=self._name,
				group=self._group,
				period=self._period,
				block=self._block,
				series=self._series,
				mass=self._mass,
				eleneg=self._eleneg,
				eleaffin=self._eleaffin,
				covrad=self._covrad,
				atmrad=self._atmrad,
				vdwrad=self._vdwrad,
				tboil=self._tboil,
				tmelt=self._tmelt,
				density=self._density,
				eleconfig=self._eleconfig,
				oxistates=self._oxistates,
				ionenergy=self._ionenergy,
				isotopes=self._isotopes,
				description=self._description
				)

	@memoized_property
	def number(self) -> int:
		"""
		The atomic number of the element.
		"""

		return self._number

	@memoized_property
	def symbol(self) -> str:
		"""
		The chemical symbol of the element.
		"""

		return self._symbol

	@memoized_property
	def name(self) -> str:
		"""
		The name of the element in English.
		"""

		return self._name

	@memoized_property
	def electrons(self) -> int:
		"""
		The number of electrons in the element.
		"""

		return self._electrons

	@memoized_property
	def protons(self) -> int:
		"""
		The number of protons in the element.
		"""

		return self._protons

	@memoized_property
	def group(self) -> int:
		"""
		The group of the element in the periodic table.
		"""

		return self._group

	@memoized_property
	def period(self) -> int:
		"""
		The Period of the element in the periodic table.
		"""

		return self._period

	@memoized_property
	def block(self) -> str:
		"""
		The Block of the element in the periodic table.
		"""

		return self._block

	@memoized_property
	def series(self) -> int:
		"""
		Index to chemical series.
		"""

		return self._series

	@memoized_property
	def mass(self) -> float:
		"""
		The relative atomic mass.

		Ratio of the average mass of atoms.
		"""
		return self._mass

	molecular_weight = mass

	@memoized_property
	def eleneg(self) -> float:
		"""
		The Electronegativity (Pauling scale).

		:rtype:

		.. latex:clearpage::
		"""

		return self._eleneg

	@memoized_property
	def eleaffin(self) -> float:
		"""
		The electron affinity in eV.
		"""

		return self._eleaffin

	@memoized_property
	def covrad(self) -> float:
		"""
		The Covalent radius in Angstrom.
		"""

		return self._covrad

	@memoized_property
	def atmrad(self) -> float:
		"""
		The Atomic radius in Angstrom.
		"""

		return self._atmrad

	@memoized_property
	def vdwrad(self) -> float:
		"""
		The Van der Waals radius in Angstrom.
		"""

		return self._vdwrad

	@memoized_property
	def tboil(self) -> float:
		"""
		The boiling temperature in K.
		"""

		return self._tboil

	@memoized_property
	def tmelt(self) -> float:
		"""
		The melting temperature in K.
		"""

		return self._tmelt

	@memoized_property
	def density(self) -> float:
		"""
		The density at 295K in g/cm³ respectively g/L.
		"""

		return self._density

	@memoized_property
	def eleconfig(self) -> str:
		"""
		The Ground state electron configuration.
		"""

		return self._eleconfig

	@memoized_property
	def oxistates(self) -> str:
		"""
		The oxidation states.
		"""

		return self._oxistates

	@memoized_property
	def ionenergy(self) -> Tuple:  # TODO
		"""
		The ionization energies in ``eV``.
		"""

		return self._ionenergy

	@memoized_property
	def isotopes(self) -> Dict[int, "Isotope"]:
		"""
		The Isotopic composition.

		* **keys**: isotope mass number
		* **values**: Isotope(relative atomic mass, abundance)

		.. latex:vspace:: -10px
		"""

		return self._isotopes

	@memoized_property
	def description(self) -> str:
		"""
		A description of the element.
		"""

		return self._description

	def __str__(self) -> str:
		return self.name

	def __repr__(self) -> str:
		ionenergy_list = []

		for i, j in enumerate(self.ionenergy):
			if i and (i % 5 == 0):
				ionenergy_list.append(f"\n        {j}")
			else:
				ionenergy_list.append(f"{j}")

		ionenergy = ", ".join(ionenergy_list)

		if len(self.ionenergy) > 5:
			ionenergy = f"(\n        {ionenergy},\n    )"
		elif len(self.ionenergy) == 1:
			ionenergy = f"({ionenergy},)"
		else:
			ionenergy = f"({ionenergy})"

		isotopes_list = []

		for massnum in sorted(self.isotopes):
			iso = self.isotopes[massnum]
			isotopes_list.append(f"{massnum}: Isotope({iso.mass}, {iso.abundance}, {massnum})")

		isotopes = ",\n        ".join(isotopes_list)

		if len(self.isotopes) > 1:
			isotopes = f"{{\n        {isotopes},\n    }},"
		else:
			isotopes = f"{{{isotopes}}},"

		return ",\n    ".join((
				f"Element(\n    {self.number}, '{self.symbol}', '{self.name}'",
				f"group={self.group}, period={self.period},"
				f" block='{self.block}', series={self.series}",
				f"mass={self.mass}, eleneg={self.eleneg},"
				f" eleaffin={self.eleaffin}",
				f"covrad={self.covrad}, atmrad={self.atmrad},"
				f" vdwrad={self.vdwrad}",
				f"tboil={self.tboil}, tmelt={self.tmelt}, density={self.density}",
				f"eleconfig='{self.eleconfig}'",
				f"oxistates='{self.oxistates}'",
				f"ionenergy={ionenergy}",
				f"isotopes={isotopes}\n)"
				))

	@memoized_property
	def nominalmass(self) -> int:
		"""
		The mass number of the most abundant natural stable isotope.
		"""

		nominalmass = 0
		maxabundance = 0

		for massnum, iso in self.isotopes.items():
			if iso.abundance > maxabundance:
				maxabundance = iso.abundance
				nominalmass = massnum

		return nominalmass

	@memoized_property
	def neutrons(self) -> int:
		"""
		The number of neutrons in the most abundant natural stable isotope.
		"""

		return self.nominalmass - self.protons

	@memoized_property
	def exactmass(self) -> float:
		"""
		The relative atomic mass calculated from the isotopic composition.
		"""

		return sum(iso.mass * iso.abundance for iso in self.isotopes.values())

	@memoized_property
	def eleconfig_dict(self) -> Dict[Tuple, int]:  # TODO
		"""
		The ground state electron configuration.

		Mapping of Tuple(shell, subshell): electrons.
		"""

		adict = {}
		if self.eleconfig.startswith('['):
			base = self.eleconfig.split(' ', 1)[0][1:-1]
			adict.update(_elements.ELEMENTS[base].eleconfig_dict)
		for e in self.eleconfig.split()[bool(adict):]:
			adict[(int(e[0]), e[1])] = int(e[2:]) if len(e) > 2 else 1
		return adict

	@memoized_property
	def eleshells(self) -> Tuple[int, ...]:
		"""
		The number of electrons per shell as tuple.
		"""

		eleshells = [0, 0, 0, 0, 0, 0, 0]
		for key, val in self.eleconfig_dict.items():
			eleshells[key[0] - 1] += val
		return tuple(ele for ele in eleshells if ele)

	def validate(self) -> None:
		"""
		Check consistency of the data.

		:raises ValueError: If there are any validation issues.
		"""

		assert self.period in _table.PERIODS
		assert self.group in _table.GROUPS
		assert self.block in _table.BLOCKS
		assert self.series in _table.SERIES

		if self.number != self.protons:
			raise ValueError(f"{self.symbol} - atomic number must equal proton number")

		if self.protons != sum(self.eleshells):
			raise ValueError(f"{self.symbol} - number of protons must equal electrons")

		if len(self.ionenergy) > 1:
			ionev_ = self.ionenergy[0]
			for ionev in self.ionenergy[1:]:
				if ionev <= ionev_:
					raise ValueError(f"{self.symbol} - ionenergy not increasing")
				ionev_ = ionev

		mass = 0.0
		frac = 0.0

		for iso in self.isotopes.values():
			mass += iso.abundance * iso.mass
			frac += iso.abundance

		if abs(mass - self.mass) > 0.03:
			raise ValueError(
					f"{self.symbol} - average of isotope masses "
					f"({mass:.4f}) != mass ({self.mass:.4f})"
					)

		if abs(frac - 1.0) > 1e-9:
			raise ValueError(f"{self.symbol} - sum of isotope abundances != 1.0")


@prettify_docstrings
class Isotope(Dictable):
	"""
	Isotope massnumber, relative atomic mass, and abundance.

	:param mass: The mass of the isotope.
	:param abundance: The natural abundance of the isotope.
	:param massnumber: The mass number of the isotope.

	.. latex:clearpage::
	"""

	def __init__(self, mass: float = 0.0, abundance: float = 1.0, massnumber: int = 0):
		super().__init__()

		self._mass: float = mass
		self._abundance: float = abundance
		self._massnumber: int = massnumber

	@memoized_property
	def mass(self) -> float:
		"""
		The mass of the isotope.
		"""

		return self._mass

	@memoized_property
	def abundance(self) -> float:
		"""
		The natural abundance of the isotope.
		"""

		return self._abundance

	@memoized_property
	def massnumber(self) -> int:
		"""
		The mass number of the isotope.
		"""

		return self._massnumber

	def __str__(self) -> str:
		return f"{self.massnumber}, {self.mass:.4f}, {self.abundance * 100:.6f}%"

	def __repr__(self) -> str:
		return f"Isotope({repr(self.mass)}, {repr(self.abundance)}, {repr(self.massnumber)})"

	@property
	def __dict__(self):  # noqa: MAN002
		return dict(
				mass=self.mass,
				abundance=self.abundance,
				massnumber=self.massnumber,
				)


# Isotope 0 Key:
# mass of the most abundant isotope and 1.0 abundance.


# TODO: make frozen
@prettify_docstrings
class Elements(Iterable[Element]):
	r"""
	Ordered dict of Elements with lookup by number, symbol, and name.

	:param \*elements: The elements to add to the dictionary.

	.. autosummary-widths:: 1/2
	.. latex:vspace:: -10px
	"""

	def __init__(self, *elements: Element):
		self._list: List[Element] = []
		self._dict: Dict[Union[str, int], Element] = {}

		for element in elements:
			if element.number > len(self._list) + 1:
				raise ValueError("Elements must be added in order")
			if element.number <= len(self._list):
				self._list[element.number - 1] = element
			else:
				self._list.append(element)

			self._dict[element.number] = element
			self._dict[element.symbol] = element
			self.add_alternate_spelling(element, element.name)

	def __str__(self) -> str:
		return f'[{", ".join(ele.symbol for ele in self._list)}]'

	def __repr__(self) -> str:
		elements = ",\n    ".join(
				"\n    ".join(line for line in repr(element).splitlines()) for element in self._list
				)
		elements = f"Elements(\n    {elements},\n)"
		return elements

	def __contains__(self, item) -> bool:  # noqa: MAN001
		return item in self._dict

	def __iter__(self) -> Iterator[Element]:
		"""
		Returns an iterator over the elements, in order.
		"""

		return iter(self._list)

	def __len__(self) -> int:
		"""
		Returns the number of elements.
		"""

		return len(self._list)

	@overload
	def __getitem__(self, key: slice) -> List[Element]: ...

	@overload
	def __getitem__(self, key: Union[str, int, float]) -> Element: ...

	def __getitem__(self, key):  # noqa: MAN001,MAN002
		"""
		Return ``self[key]``.

		:param key: If a string, return the :class:`~.Element` with that name or symbol.
			If a number, return the element with that atomic number.
		"""

		# TODO: slice docstring

		if isinstance(key, str):
			try:
				return self._dict[key.casefold()]
			except KeyError:
				return self._dict[key]
		elif isinstance(key, int):
			return self._dict[key]
		elif isinstance(key, float):
			return self._dict[int(key)]
		elif isinstance(key, slice):
			start, stop, step = key.indices(len(self._list))
			return self._list[slice(start - 1, stop - 1, step)]
		else:
			try:
				symbol, isotope = self.split_isotope(key)
				return self._dict[symbol.capitalize()]
			except (ValueError, KeyError):
				raise KeyError(f"Unknown key: '{key}'")

	@functools.lru_cache()
	def split_isotope(self, string: str) -> Tuple[str, int]:
		"""
		Returns the symbol and mass number for the isotope represented by ``string``.

		Valid isotopes include ``'[C12]'``, ``'C[12]'`` and ``'[12C]'``.

		:param string:

		:return: Tuple representing the element and the isotope number.
		"""

		# this package
		from chemistry_tools.formulae.utils import split_isotope

		return split_isotope(string)

	def add_alternate_spelling(self, element: Element, spelling: str) -> None:
		"""
		Adds an alternate spelling for an element.

		:param element:
		:param spelling:
		"""

		self._dict[spelling] = element
		self._dict[spelling.lower()] = element
		self._dict[spelling.casefold()] = element

	@memoized_property
	def symbols(self) -> List[str]:
		"""
		The symbols of the elements.
		"""

		return [element.symbol for element in sorted(self._list, key=lambda e: e.number)]

	@memoized_property
	def names(self) -> List[str]:
		"""
		The names of the elements.
		"""

		return [str(element) for element in sorted(self._list, key=lambda e: e.number)]

	@memoized_property
	def lower_names(self) -> List[str]:
		"""
		The names of the elements, all in lowercase.
		"""

		return [str(element).lower() for element in sorted(self._list, key=lambda e: e.number)]


@doctools.append_docstring_from(Element)
class HeavyHydrogen(Element):
	"""
	Subclass of :class:`~.Element` to handle the Heavy Hydrogen isotopes Deuterium and Tritium.
	"""

	@functools.wraps(Element.__init__)
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

		if self.symbol not in {'D', 'T'}:
			raise ValueError("'HeavyHydrogen' can only be used for Deuterium and Tritium")

	@memoized_property
	def nominalmass(self) -> int:
		"""
		Return mass number of most abundant natural stable isotope.
		"""

		if self.symbol == 'D':
			return 2
		elif self.symbol == 'T':
			return 3
		else:
			raise ValueError("Unknown heavy hydrogen isotope.")

	@memoized_property
	def as_isotope(self) -> str:
		"""
		Return the isotope in ``H[X]`` format.
		"""

		return f"H[{self.nominalmass}]"



#!/usr/bin/env python3
#
#  _elements.py
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on molmass (https://github.com/cgohlke/molmass)
#  |  Copyright (c) 1990-2020, Christoph Gohlke
#  |  All rights reserved.
#  |  Licensed under the BSD 3-Clause License
#  |  Redistribution and use in source and binary forms, with or without
#  |  modification, are permitted provided that the following conditions are met:
#  |
#  |  1. Redistributions of source code must retain the above copyright notice,
#  |     this list of conditions and the following disclaimer.
#  |
#  |  2. Redistributions in binary form must reproduce the above copyright notice,
#  |     this list of conditions and the following disclaimer in the documentation
#  |     and/or other materials provided with the distribution.
#  |
#  |  3. Neither the name of the copyright holder nor the names of its
#  |     contributors may be used to endorse or promote products derived from
#  |     this software without specific prior written permission.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  |  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  |  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  |  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  |  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  |  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  |  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  |  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  |  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  |  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  |  POSSIBILITY OF SUCH DAMAGE.
#  |
#

# stdlib
from typing import Dict

# this package
from . import classes
from .actinides import *
from .alkali_metals import *
from .alkaline_earth_metals import *
from .chalcogens import *
from .halogens import *
from .lanthanides import *
from .noble_gases import *
from .pnictogens import *
from .tetrels import *
from .transition_metals import *
from .triels import *

__all__ = ("ELEMENTS", 'H', 'D', 'T')

# TODO: nominal mass (massnumber) and monoisotopic mass

H = classes.Element(
		1,
		'H',
		"Hydrogen",
		group=1,
		period=1,
		block='s',
		series=1,
		mass=1.007941,
		eleneg=2.2,
		eleaffin=0.75420375,
		covrad=0.32,
		atmrad=0.79,
		vdwrad=1.2,
		tboil=20.28,
		tmelt=13.81,
		density=0.084,
		eleconfig="1s",
		oxistates="1*, -1",
		ionenergy=(13.5984, ),
		isotopes={
				1: (1.00782503207, 0.999885),
				2: (2.0141017778, 0.000115),
				3: (3.0160492777, 0.0),
				4: (4.02781, 0.0),
				5: (5.03531, 0.0),
				6: (6.04494, 0.0),
				7: (7.05275, 0.0),
				},
		description=(
				"Colourless, odourless gaseous chemical element. Lightest and most "
				"abundant element in the universe. Present in water and in all "
				"organic compounds. Chemically reacts with most elements. Discovered "
				"by Henry Cavendish in 1776."
				)
		)

D = classes.HeavyHydrogen(
		1,
		'D',
		"Deuterium",
		mass=2.0141017778,
		isotopes={
				1: (1.00782503207, 0.999885),
				2: (2.0141017778, 0.000115),
				3: (3.0160492777, 0.0),
				4: (4.02781, 0.0),
				5: (5.03531, 0.0),
				6: (6.04494, 0.0),
				7: (7.05275, 0.0),
				},
		)

T = classes.HeavyHydrogen(
		1,
		'T',
		"Tritium",
		mass=3.0160492777,
		isotopes={
				1: (1.00782503207, 0.999885),
				2: (2.0141017778, 0.000115),
				3: (3.0160492777, 0.0),
				4: (4.02781, 0.0),
				5: (5.03531, 0.0),
				6: (6.04494, 0.0),
				7: (7.05275, 0.0),
				},
		)

ELEMENTS = classes.Elements(
		H,
		He,
		Li,
		Be,
		B,
		C,
		N,
		O,
		F,
		Ne,
		Na,
		Mg,
		Al,
		Si,
		P,
		S,
		Cl,
		Ar,
		K,
		Ca,
		Sc,
		Ti,
		V,
		Cr,
		Mn,
		Fe,
		Co,
		Ni,
		Cu,
		Zn,
		Ga,
		Ge,
		As,
		Se,
		Br,
		Kr,
		Rb,
		Sr,
		Y,
		Zr,
		Nb,
		Mo,
		Tc,
		Ru,
		Rh,
		Pd,
		Ag,
		Cd,
		In,
		Sn,
		Sb,
		Te,
		I,
		Xe,
		Cs,
		Ba,
		La,
		Ce,
		Pr,
		Nd,
		Pm,
		Sm,
		Eu,
		Gd,
		Tb,
		Dy,
		Ho,
		Er,
		Tm,
		Yb,
		Lu,
		Hf,
		Ta,
		W,
		Re,
		Os,
		Ir,
		Pt,
		Au,
		Hg,
		Tl,
		Pb,
		Bi,
		Po,
		At,
		Rn,
		Fr,
		Ra,
		Ac,
		Th,
		Pa,
		U,
		Np,
		Pu,
		Am,
		Cm,
		Bk,
		Cf,
		Es,
		Fm,
		Md,
		No,
		Lr,
		Rf,
		Db,
		Sg,
		Bh,
		Hs,
		Mt,
		Ds,
		Rg,
		Cn,
		Nh,
		Fl,
		Mc,
		Lv,
		Ts,
		Og,
		)

# ELEMENTS = Elements(
# H,                                                                                                                          He,
# Li, Be,                                                                                                 B,  C,  N,  O,  F,  Ne,
# Na, Mg,                                                                                                 Al, Si, P,  S,  Cl, Ar,
# K,  Ca,                                                         Sc, Ti, V,  Cr, Mn, Fe, Co, Ni, Cu, Zn, Ga, Ge, As, Se, Br, Kr,
# Rb, Sr,                                                         Y,  Zr, Nb, Mo, Tc, Ru, Rh, Pd, Ag, Cd, In, Sn, Sb, Te, I,  Xe,
# Cs, Ba, La, Ce, Pr, Nd, Pm, Sm, Eu, Gd, Tb, Dy, Ho, Er, Tm, Yb, Lu, Hf, Ta, W,  Re, Os, Ir, Pt, Au, Hg, Tl, Pb, Bi, Po, At, Rn,
# Fr, Ra, Ac, Th, Pa, U,  Np, Pu, Am, Cm, Bk, Cf, Es, Fm, Md, No, Lr, Rf, Db, Sg, Bh, Hs, Mt, Ds, Rg, Cn, Nh, Fl, Mc, Lv, Ts, Og,
# )

alternate_spellings: Dict[str, str] = {
		"Caesium": "Cesium",  # AmE
		"Aluminium": "Aluminum",  # AmE
		"Sulfur": "Sulphur",  # Not IUPAC and not etymologically correct, but widely used in BrE
		}

for symbol, spelling in alternate_spellings.items():
	ELEMENTS.add_alternate_spelling(ELEMENTS[symbol], spelling)

for element in [D, T]:
	ELEMENTS.add_alternate_spelling(element, element.name)
	ELEMENTS._dict[element.symbol] = element
	# ELEMENTS._list.append(element)



#!/usr/bin/env python3
#
#  spectrum_similarity.py
"""
Mass spectrum similarity calculations.
"""
#
#  Copyright (c) 2019-2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Adapted from SpectrumSimilarity.R
#  Part of OrgMassSpecR
#  |  Copyright (c) 2011-2017 Nathan Dodder <nathand@sccwrp.org>
#  |  Available under the BSD 2-Clause License
#  |
#  |  Redistribution and use in source and binary forms, with or without modification,
#  |  are permitted provided that the following conditions are met:
#  |
#  |    Redistributions of source code must retain the above copyright notice, this
#  |    list of conditions and the following disclaimer.
#  |
#  |    Redistributions in binary form must reproduce the above copyright notice, this
#  |    list of conditions and the following disclaimer in the documentation and/or
#  |    other materials provided with the distribution.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#  |  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#  |  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  |  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
#  |  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#  |  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  |  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
#  |  ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  |  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  |  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

# stdlib
from typing import Mapping, Optional, Sequence, Tuple, Union

# 3rd party
import numpy
import pandas  # type: ignore[import]

__all__ = ["spectrum_similarity", "normalize", "create_array", "SpectrumSimilarity"]


class SpectrumSimilarity:
	"""
	Calculate the similarity score for two mass spectra.

	:param spec_top: Array containing the experimental spectrum's peak list with the *m/z* values in the
		first column and corresponding intensities in the second
	:param spec_bottom: Array containing the reference spectrum's peak list with the *m/z* values in the
		first column and corresponding intensities in the second
	:param b: numeric value specifying the baseline threshold for peak identification.
		Expressed as a percent of the maximum intensity.
	:param xlim: tuple of length 2, defining the beginning and ending values of the x-axis.

	.. versionadded:: 1.0.0

	.. TODO:
		x_threshold: numeric value specifying
		t: numeric value specifying the tolerance used to align the *m/z* values of the two spectra.
	"""

	top_df: pandas.DataFrame
	_top_df_plot: pandas.DataFrame  # includes peaks below ``b``
	bottom_df: pandas.DataFrame
	_bottom_df_plot: pandas.DataFrame  # includes peaks below ``b``
	b: float
	alignment: pandas.DataFrame

	def __init__(
			self,
			spec_top: numpy.ndarray,
			spec_bottom: numpy.ndarray,
			# t: float = 0.25,
			b: float = 1,
			xlim: Tuple[int, int] = (50, 1200),  # x_threshold: float = 0,
			):

		# if x_threshold < 0:
		# 	raise ValueError("x_threshold argument must be zero or a positive number")

		self.b = b
		self.xlim = xlim

		# format spectra and normalize intensitites
		self.top_df, self._top_df_plot = self._build_dataframe(spec_top)
		self.bottom_df, self._bottom_df_plot = self._build_dataframe(spec_bottom)

		# align the m/z axis of the two spectra, the bottom spectrum is used as the reference

		# Unimplemented R code
		#   for(i in 1:nrow(bottom))
		# 	top["mz"][bottom["mz"][i] >= top["mz"] - t & bottom["mz"][i] <= top["mz"] + t] = bottom["mz"][i]
		# 	top[,1][bottom[,1][i] >= top[,1] - t & bottom[,1][i] <= top[,1] + t] <- bottom[,1][i]
		#   alignment <- merge(top, bottom, by = 1, all = TRUE)
		#   if(length(unique(alignment[,1])) != length(alignment[,1])) warning("the m/z tolerance is set too high")
		# alignment[,c(2,3)][is.na(alignment[,c(2,3)])] <- 0   # convert NAs to zero (R-Help, Sept. 15, 2004, John Fox)
		# names(alignment) <- c("mz", "intensity.top", "intensity.bottom")

		alignment = pandas.merge(self.top_df, self.bottom_df, on="mz", how="outer")
		self.alignment = alignment.fillna(value=0)  # Convert NaN to 0
		self.alignment.columns = ["mz", "intensity_top", "intensity_bottom"]

		reverse_alignment = pandas.merge(self.top_df, self.bottom_df, on="mz", how="right")
		self.reverse_alignment = reverse_alignment.dropna()  # Remove rows containing NaN
		self.reverse_alignment.columns = ["mz", "intensity_top", "intensity_bottom"]

	def _calculate_score(self, alignment: pandas.DataFrame) -> float:
		u = numpy.array(alignment.iloc[:, 1])
		v = numpy.array(alignment.iloc[:, 2])

		return numpy.dot(u, v) / (numpy.sqrt(numpy.sum(numpy.square(u))) * numpy.sqrt(numpy.sum(numpy.square(v))))

	def score(self) -> Tuple[float, float]:
		"""
		Returns the similarity score.
		"""

		# Unimplemented R code
		# alignment <- alignment[alignment[,1] >= x.threshold, ]

		similarity_score = self._calculate_score(self.alignment)
		reverse_similarity_score = self._calculate_score(self.reverse_alignment)

		return similarity_score, reverse_similarity_score

	def plot(
			self,
			top_label: Optional[str] = None,
			bottom_label: Optional[str] = None,
			filter: bool = False,  # Whether peaks below b should be filtered  # noqa: A002  # pylint: disable=redefined-builtin
			) -> None:
		"""
		Plot the mass spectra head to tail.

		:param top_label: string to label the top spectrum.
		:param bottom_label: string to label the bottom spectrum.
		"""

		# 3rd party
		import matplotlib.pyplot as plt  # type: ignore[import]  # nodep

		_, ax = plt.subplots()
		# fig.scatter(top_plot["mz"],top_plot["intensity"], s=0)

		if filter:
			ax.vlines(self.top_df["mz"], 0, self.top_df["intensity"], color="blue")
			ax.vlines(self.bottom_df["mz"], 0, -self.bottom_df["intensity"], color="red")
		else:
			ax.vlines(self._top_df_plot["mz"], 0, self._top_df_plot["intensity"], color="blue")
			ax.vlines(self._bottom_df_plot["mz"], 0, -self._bottom_df_plot["intensity"], color="red")

		ax.set_ylim(-125, 125)
		ax.set_xlim(self.xlim[0], self.xlim[1])
		ax.axhline(color="black", linewidth=0.5)
		ax.set_ylabel("Intensity (%)")
		ax.set_xlabel("m/z", style="italic", family="serif")

		h_centre = self.xlim[0] + (self.xlim[1] - self.xlim[0]) // 2

		ax.text(h_centre, 110, top_label, horizontalalignment="center", verticalalignment="center")
		ax.text(h_centre, -110, bottom_label, horizontalalignment="center", verticalalignment="center")
		return ax

		# Unimplemented R code
		# 	ticks <- c(-100, -50, 0, 50, 100)
		# 	plot.window(xlim = c(0, 20), ylim = c(-10, 10))
		#
		#
		#   if(output.list == TRUE) {
		#
		# 	# Grid graphics head to tail plot
		#
		# 	headTailPlot <- function() {
		#
		# 	  pushViewport(plotViewport(c(5, 5, 2, 2)))
		# 	  pushViewport(dataViewport(xscale = xlim, yscale = c(-125, 125)))
		#
		# 	  grid.rect()
		# 	  tmp <- pretty(xlim)
		# 	  xlabels <- tmp[tmp >= xlim[1] & tmp <= xlim[2]]
		# 	  grid.xaxis(at = xlabels)
		# 	  grid.yaxis(at = c(-100, -50, 0, 50, 100))
		#
		# 	  grid.segments(top_plot$mz,
		# 					top_plot$intensity,
		# 					top_plot$mz,
		# 					rep(0, length(top_plot$intensity)),
		# 					default.units = "native",
		# 					gp = gpar(lwd = 0.75, col = "blue"))
		#
		# 	  grid.segments(bottom_plot$mz,
		# 					-bottom_plot$intensity,
		# 					bottom_plot$mz,
		# 					rep(0, length(bottom_plot$intensity)),
		# 					default.units = "native",
		# 					gp = gpar(lwd = 0.75, col = "red"))
		#
		# 	  grid.abline(intercept = 0, slope = 0)
		#
		# 	  grid.text("intensity (%)", x = unit(-3.5, "lines"), rot = 90)
		# 	  grid.text("m/z", y = unit(-3.5, "lines"))
		#
		# 	  popViewport(1)
		# 	  pushViewport(dataViewport(xscale = c(0, 20), yscale = c(-10, 10)))
		# 	  grid.text(top.label, unit(10, "native"), unit(9, "native"))
		# 	  grid.text(bottom.label, unit(10, "native"), unit(-9, "native"))
		#
		# 	  popViewport(2)
		#
		# 	}
		#
		# 	p <- grid.grabExpr(headTailPlot())
		#
		#   }
		#
		# with pandas.option_context('display.max_rows', None, 'display.max_columns', None):
		# 	print(similarity_score)

		# Unimplemented R code
		#
		# return(list(similarity.score = similarity_score,
		# 	alignment = alignment,
		# 	plot = p))
		#

	def print_alignment(self) -> None:
		"""
		Print the dataframe giving aligned peaks in the top and bottom spectra.
		"""

		with pandas.option_context("display.max_rows", None, "display.max_columns", None):
			print(self.alignment)

	def _build_dataframe(self, spectrum: numpy.ndarray) -> Tuple[pandas.DataFrame, pandas.DataFrame]:
		# format spectra and normalize intensitites
		tmp = pandas.DataFrame(data=spectrum, columns=["mz", "intensity"])
		tmp["normalized"] = tmp.apply(normalize, args=(max(tmp["intensity"]), ), axis=1)
		tmp = tmp[tmp["mz"].between(self.xlim[0], self.xlim[1])]
		plot = tmp[["mz", "normalized"]].copy()  # data frame for plotting spectrum
		plot.columns = ["mz", "intensity"]
		out = plot[plot["intensity"] >= self.b]  # data frame for similarity score calculation

		return out, plot


# Deprecated
def spectrum_similarity(
		spec_top: numpy.ndarray,
		spec_bottom: numpy.ndarray,
		t: float = 0.25,
		b: float = 10,
		top_label: Optional[str] = None,
		bottom_label: Optional[str] = None,
		xlim: Tuple[int, int] = (50, 1200),
		x_threshold: float = 0,
		print_alignment: bool = False,
		print_graphic: bool = True,
		output_list: bool = False,
		) -> Union[Tuple[float, float], Tuple[float, float, pandas.DataFrame]]:
	"""
	Calculate the similarity score for two mass spectra.

	.. attention:: The :class:`~.SpectrumSimilarity` class is recommended over this function.

	:param spec_top: Array containing the experimental spectrum's peak list with the *m/z* values in the
		first column and corresponding intensities in the second
	:param spec_bottom: Array containing the reference spectrum's peak list with the *m/z* values in the
		first column and corresponding intensities in the second
	:param t: numeric value specifying the tolerance used to align the *m/z* values of the two spectra.
	:param b: numeric value specifying the baseline threshold for peak identification.
		Expressed as a percent of the maximum intensity.
	:param top_label: string to label the top spectrum.
	:param bottom_label: string to label the bottom spectrum.
	:param xlim: tuple of length 2, defining the beginning and ending values of the x-axis.
	:param x_threshold:
	:param print_alignment:  whether the intensities should be printed
	:param print_graphic:
	:param output_list: whether the intensities should be returned as a third element of the tuple.
	"""

	# format spectra and normalize intensitites
	top_tmp = pandas.DataFrame(data=spec_top, columns=["mz", "intensity"])
	top_tmp["normalized"] = top_tmp.apply(normalize, args=(max(top_tmp["intensity"]), ), axis=1)
	top_tmp = top_tmp[top_tmp["mz"].between(xlim[0], xlim[1])]
	top_plot = top_tmp[["mz", "normalized"]].copy()  # data frame for plotting spectrum
	top_plot.columns = ["mz", "intensity"]
	top = top_plot[top_plot["intensity"] >= b]  # data frame for similarity score calculation

	bottom_tmp = pandas.DataFrame(data=spec_bottom, columns=["mz", "intensity"])
	bottom_tmp["normalized"] = bottom_tmp.apply(normalize, args=(max(bottom_tmp["intensity"]), ), axis=1)
	bottom_tmp = bottom_tmp[bottom_tmp["mz"].between(xlim[0], xlim[1])]
	bottom_plot = bottom_tmp[["mz", "normalized"]].copy()  # data frame for plotting spectrum
	bottom_plot.columns = ["mz", "intensity"]
	bottom = bottom_plot[bottom_plot["intensity"] >= b]  # data frame for similarity score calculation

	# align the m/z axis of the two spectra, the bottom spectrum is used as the reference

	# Unimplemented R code
	#   for(i in 1:nrow(bottom))
	# 	top["mz"][bottom["mz"][i] >= top["mz"] - t & bottom["mz"][i] <= top["mz"] + t] = bottom["mz"][i]
	# 	top[,1][bottom[,1][i] >= top[,1] - t & bottom[,1][i] <= top[,1] + t] <- bottom[,1][i]
	#   alignment <- merge(top, bottom, by = 1, all = TRUE)
	#   if(length(unique(alignment[,1])) != length(alignment[,1])) warning("the m/z tolerance is set too high")
	# alignment[,c(2,3)][is.na(alignment[,c(2,3)])] <- 0   # convert NAs to zero (R-Help, Sept. 15, 2004, John Fox)
	# names(alignment) <- c("mz", "intensity.top", "intensity.bottom")
	#
	alignment = pandas.merge(top, bottom, on="mz", how="outer")
	alignment.fillna(value=0, inplace=True)  # Convert NaN to 0
	alignment.columns = ["mz", "intensity_top", "intensity_bottom"]
	if print_alignment:
		with pandas.option_context("display.max_rows", None, "display.max_columns", None):
			print(alignment)

	# similarity score calculation

	if x_threshold < 0:
		raise ValueError("x_threshold argument must be zero or a positive number")

	# Unimplemented R code
	# alignment <- alignment[alignment[,1] >= x.threshold, ]

	u = numpy.array(alignment.iloc[:, 1])
	v = numpy.array(alignment.iloc[:, 2])

	similarity_score = numpy.dot(u, v) / (
			numpy.sqrt(numpy.sum(numpy.square(u))) * numpy.sqrt(numpy.sum(numpy.square(v)))
			)

	# Reverse Match
	reverse_alignment = pandas.merge(top, bottom, on="mz", how="right")
	reverse_alignment = reverse_alignment.dropna()  # Remove rows containing NaN
	reverse_alignment.columns = ["mz", "intensity_top", "intensity_bottom"]
	u = numpy.array(reverse_alignment.iloc[:, 1])
	v = numpy.array(reverse_alignment.iloc[:, 2])

	reverse_similarity_score = numpy.dot(u, v) / (
			numpy.sqrt(numpy.sum(numpy.square(u))) * numpy.sqrt(numpy.sum(numpy.square(v)))
			)

	# generate plot

	if print_graphic:
		# 3rd party
		import matplotlib.pyplot as plt  # nodep

		fig, ax = plt.subplots()
		# fig.scatter(top_plot["mz"],top_plot["intensity"], s=0)
		ax.vlines(top_plot["mz"], 0, top_plot["intensity"], color="blue")
		ax.vlines(bottom["mz"], 0, -bottom["intensity"], color="red")
		ax.set_ylim(-125, 125)
		ax.set_xlim(xlim[0], xlim[1])
		ax.axhline(color="black", linewidth=0.5)
		ax.set_ylabel("Intensity (%)")
		ax.set_xlabel("m/z", style="italic", family="serif")

		h_centre = xlim[0] + (xlim[1] - xlim[0]) // 2

		ax.text(h_centre, 110, top_label, horizontalalignment="center", verticalalignment="center")
		ax.text(h_centre, -110, bottom_label, horizontalalignment="center", verticalalignment="center")
		plt.show()

	# Unimplemented R code
	# 	ticks <- c(-100, -50, 0, 50, 100)
	# 	plot.window(xlim = c(0, 20), ylim = c(-10, 10))
	#
	#
	#   if(output.list == TRUE) {
	#
	# 	# Grid graphics head to tail plot
	#
	# 	headTailPlot <- function() {
	#
	# 	  pushViewport(plotViewport(c(5, 5, 2, 2)))
	# 	  pushViewport(dataViewport(xscale = xlim, yscale = c(-125, 125)))
	#
	# 	  grid.rect()
	# 	  tmp <- pretty(xlim)
	# 	  xlabels <- tmp[tmp >= xlim[1] & tmp <= xlim[2]]
	# 	  grid.xaxis(at = xlabels)
	# 	  grid.yaxis(at = c(-100, -50, 0, 50, 100))
	#
	# 	  grid.segments(top_plot$mz,
	# 					top_plot$intensity,
	# 					top_plot$mz,
	# 					rep(0, length(top_plot$intensity)),
	# 					default.units = "native",
	# 					gp = gpar(lwd = 0.75, col = "blue"))
	#
	# 	  grid.segments(bottom_plot$mz,
	# 					-bottom_plot$intensity,
	# 					bottom_plot$mz,
	# 					rep(0, length(bottom_plot$intensity)),
	# 					default.units = "native",
	# 					gp = gpar(lwd = 0.75, col = "red"))
	#
	# 	  grid.abline(intercept = 0, slope = 0)
	#
	# 	  grid.text("intensity (%)", x = unit(-3.5, "lines"), rot = 90)
	# 	  grid.text("m/z", y = unit(-3.5, "lines"))
	#
	# 	  popViewport(1)
	# 	  pushViewport(dataViewport(xscale = c(0, 20), yscale = c(-10, 10)))
	# 	  grid.text(top.label, unit(10, "native"), unit(9, "native"))
	# 	  grid.text(bottom.label, unit(10, "native"), unit(-9, "native"))
	#
	# 	  popViewport(2)
	#
	# 	}
	#
	# 	p <- grid.grabExpr(headTailPlot())
	#
	#   }
	#
	# with pandas.option_context('display.max_rows', None, 'display.max_columns', None):
	# 	print(similarity_score)

	if output_list:
		return similarity_score, reverse_similarity_score, alignment
	# Unimplemented R code
	#
	# return(list(similarity.score = similarity_score,
	# 	alignment = alignment,
	# 	plot = p))
	#
	else:
		return similarity_score, reverse_similarity_score


# simscore <- as.vector((u %*% v)^2 / (sum(u^2) * sum(v^2)))   # cos squared

# SpectrumSimilarity = spectrum_similarity


def normalize(row: Union[Mapping, pandas.Series], max_val: Union[float, str]) -> float:
	"""
	Returns the normalised intensity for each rows of a :class:`pandas.DataFrame`.

	:param row:
	:param max_val:
	"""

	# http://jonathansoma.com/lede/foundations/classes/pandas%20columns%20and%20functions/apply-a-function-to-every-row-in-a-pandas-dataframe/
	return (row["intensity"] / float(max_val)) * 100.0


def create_array(intensities: Sequence[float], mz: Sequence[float]) -> numpy.ndarray:
	"""
	Create a :class:`numpy.ndarray`, in a format appropriate for :class:`~.SpectrumSimilarity`,
	from a list of intensities and a list of *m/z* values.

	:param intensities: List of intensities
	:param mz: List of *m/z* values.
	"""  # noqa: D400

	return numpy.column_stack((mz, intensities))



#!/usr/bin/env python3
#
#  constants.py
"""
Scientific constants.
"""
#
#  Copyright (c) 2019-2021 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on ChemPy (https://github.com/bjodah/chempy)
#  |  Copyright (c) 2015-2018, Björn Dahlgren
#  |  All rights reserved.
#  |
#  |  Redistribution and use in source and binary forms, with or without modification,
#  |  are permitted provided that the following conditions are met:
#  |
#  |    Redistributions of source code must retain the above copyright notice, this
#  |    list of conditions and the following disclaimer.
#  |
#  |    Redistributions in binary form must reproduce the above copyright notice, this
#  |    list of conditions and the following disclaimer in the documentation and/or
#  |    other materials provided with the distribution.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#  |  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#  |  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  |  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
#  |  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#  |  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  |  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
#  |  ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  |  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  |  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

# stdlib
from typing import Dict, NamedTuple, Optional

# 3rd party
import quantities  # type: ignore[import]

__all__ = [
		"Constant",
		"avogadro_number",
		"plancks_constant",
		"speed_of_light",
		"electron_radius",
		"neutron_mass",
		"atomic_mass_constant",
		"faraday_constant",
		"vacuum_permittivity",
		"boltzmann_constant",
		"molar_gas_constant",
		"prefixes",
		]


_anions = {  # Incomplete
		"F-": "fluoride",
		"Cl-": "chloride",
		"Br-": "bromide",
		"I-": "iodide",
		"OH-": "hydroxide",
		"CN-": "cyanide",
		"SCN-": "thiocyanate",
		"CO3-2": "carbonate",
		"C2O4-2": "oxalate",
		"HCO3-": "hydrogencarbonate",
		"NO3-": "nitrate",
		"NO2-": "nitrite",
		"PO4-3": "phospahte",
		"HPO4-2": "hydrogenphospahte",
		"H2PO4-": "dihydrogenphospahte",
		"P-3": "phosphide",
		"SO4-2": "sulphate",
		"HSO4-": "hydrogensulphate",
		"SO3-2": "sulphite",
		"HSO3-": "hydrogensulphite",
		"S-2": "sulfide",
		"ClO-": "hypochlorite",
		"ClO2-": "chlorite",
		"ClO3-": "chlorate",
		"ClO4-": "perchlorate",
		"CrO4-2": "chromate(VI)",
		"Cr2O7-2": "dichromate(VI)",
		"MnO4-2": "manganate(VI)",
		"MnO4-": "permanganate(VII)",
		"FeO4-2": "ferrate(VI)",
		"OsO4-2": "osmate(VI)",
		"Bo3-3": "borate",
		"BiO3-": "bismuthate(V)",
		}

_cations = {  # Incomplete
		"H3O+": "hydronium",
		}

_cation_oxidation_states = {  # This needs to be reviewed, just from the top of my head
		"Cr": (2, 3),
		"Fe": (2, 3),
		"Mn": (2,),
		"Co": (2, 3),
		"Ni": (2, 3),
		"Cu": (1, 2, 3),
		"Ag": (1, 2),
		"Au": (3,),
		"Zn": (2,),
		"Cd": (2,),
		"Hg": (1, 2),  # Tricky: Hg2+2
		"Al": (3,),
		"Ga": (3,),
		"In": (3,),
		"Tl": (1, 3),
		"Sn": (2, 4),
		"Pb": (2, 4),
		"Bi": (3,),
		"Sb": (3,),
		}


class Constant(NamedTuple):
	"""
	Represents a scientific constant.
	"""

	#: The name of the constant.
	name: str

	#: The value of the constant.
	value: float

	#: The constant's unit.
	unit: quantities.quantity.Quantity

	#: An optional symbol for the constant. Default :py:obj:`None`.
	symbol: Optional[str] = None

	def as_quantity(self) -> quantities.quantity.Quantity:
		"""
		Returns the constant as a :class:`quantities.quantity.Quantity` object.
		"""

		return self.value * self.unit

	def __float__(self) -> float:
		"""
		Returns the constant as a float (without the unit).
		"""

		return float(self.value)

	def __int__(self) -> int:
		"""
		Returns the constant as an integer (without the unit).
		"""

		return int(self.value)


# The following from periodictable
# Public Domain data
# Author: Paul Kienzle

#: Avogadro's constant (Avogadro's number)
avogadro_number = avogadro_constant = Constant(
		name="Avogadro constant", value=6.02214179e23, unit=1 / quantities.mol, symbol="N<sub>A</sub>"
		)  # (30)

#: Planck's constant
plancks_constant = planck_constant = Constant(
		name="Planck's constant",
		value=4.13566733e-15 * (10**34),
		unit=quantities.electron_volt / quantities.second,
		symbol='h'
		)  # (10)

#: The speed of light in a vacuum.
speed_of_light = Constant(
		name="Speed of light", value=299792458, unit=quantities.m / quantities.second, symbol='c'
		)  # (exact)

#: Electron Radius
electron_radius = Constant(name="Electron radius", value=2.8179402894e-15, unit=quantities.m, symbol="rₑ")  # (58)

# From NIST Reference on Constants, Units, and Uncertainty
#   http://physics.nist.gov/cuu/index.html
# neutron mass = 1.008 664 915 97(43) u
# atomic mass constant m_u = 1.660 538 782(83) x 10-27 kg

#: Neutron mass
neutron_mass = Constant(
		name="Neutron mass", value=1.00866491597, unit=quantities.atomic_mass_unit, symbol="n<sup>o</sup>"
		)  # (43)

#: The atomic mass constant.
atomic_mass_constant = float(quantities.atomic_mass_unit.rescale(quantities.kg))

#: Faraday constant
faraday_constant = Constant(
		name="Faraday constant",
		value=96485.3321233100184,
		unit=quantities.coulomb * (1 / quantities.mol),
		symbol='F'
		)

#: Vacuum permittivity
vacuum_permittivity = Constant(
		"Vacuum permittivity", value=8.8541878128e-12, unit=quantities.farad / quantities.metre, symbol="ε₀"
		)

#: Boltzmann constant
boltzmann_constant = Constant(
		name="Boltzmann constant",
		value=1.380649e-23,
		unit=quantities.joule / quantities.kelvin,
		symbol="k<sub>B</sub>"
		)

#: Molar gas constant
molar_gas_constant = Constant(
		name="Molar gas constant",
		value=8.31446261815324,
		unit=quantities.joule / quantities.kelvin / quantities.mol,
		symbol='R'
		)

#: Numerical IUPAC prefixes (e.g. **mono-**).
prefixes: Dict[int, str] = {
		1: "mono",
		2: "di",
		3: "tri",
		4: "tetra",
		5: "penta",
		6: "hexa",
		7: "hepta",
		8: "octa",
		9: "nona",
		10: "deca",
		11: "undeca",
		12: "dodeca",
		13: "trideca",
		14: "tetradeca",
		15: "pentadeca",
		16: "hexadeca",
		17: "heptadeca",
		18: "octadeca",
		19: "nonadeca",
		20: "icosa",
		21: "henicosa",
		22: "docosa",
		23: "tricosa",
		30: "triaconta",
		31: "hentriaconta",
		32: "dotriaconta",
		40: "tetraconta",
		50: "pentaconta",
		60: "hexaconta",
		70: "heptaconta",
		80: "octaconta",
		90: "nonaconta",
		100: "hecta",
		200: "dicta",
		300: "tricta",
		400: "tetracta",
		500: "pentacta",
		600: "hexacta",
		700: "heptacta",
		800: "octacta",
		900: "nonacta",
		1000: "kilia",
		2000: "dilia",
		3000: "trilia",
		4000: "tetralia",
		5000: "pentalia",
		6000: "hexalia",
		7000: "heptalia",
		8000: "octalia",
		9000: "nonalia",
		}
